import json
import argparse

from starfish.image import ImageStack
from starfish.pipeline.pipelinecomponent import PipelineComponent
from starfish.util.argparse import FsExistsType
from . import watershed
from ._base import SegmentationAlgorithmBase


class Segmentation(PipelineComponent):

    segmentation_group: argparse.ArgumentParser

    @classmethod
    def implementing_algorithms(cls):
        return SegmentationAlgorithmBase.__subclasses__()

    @classmethod
    def add_to_parser(cls, subparsers):
        """Adds the segmentation component to the CLI argument parser."""
        segmentation_group = subparsers.add_parser("segment")
        segmentation_group.add_argument("--hybridization-stack", type=FsExistsType(), required=True)
        segmentation_group.add_argument("--nuclei-stack", type=FsExistsType(), required=True)
        segmentation_group.add_argument("-o", "--output", required=True)
        segmentation_group.set_defaults(starfish_command=Segmentation._cli)
        segmentation_subparsers = segmentation_group.add_subparsers(dest="segmentation_algorithm_class")

        for algorithm_cls in cls.algorithm_to_class_map().values():
            group_parser = segmentation_subparsers.add_parser(algorithm_cls.get_algorithm_name())
            group_parser.set_defaults(segmentation_algorithm_class=algorithm_cls)
            algorithm_cls.add_arguments(group_parser)

        cls.segmentation_group = segmentation_group

    @classmethod
    def _cli(cls, args, print_help=False):
        """Runs the segmentation component based on parsed arguments."""
        if args.segmentation_algorithm_class is None or print_help:
            cls.segmentation_group.print_help()
            cls.segmentation_group.exit(status=2)

        instance = args.segmentation_algorithm_class(**vars(args))

        print('Segmenting ...')
        hybridization_stack = ImageStack.from_path_or_url(args.hybridization_stack)
        nuclei_stack = ImageStack.from_path_or_url(args.nuclei_stack)
        regions = instance.segment(hybridization_stack, nuclei_stack)
        geojson = regions_to_geojson(regions, use_hull=False)

        print("Writing | regions geojson to: {}".format(args.output))
        with open(args.output, "w") as f:
            f.write(json.dumps(geojson))


def regions_to_geojson(r, use_hull=True):
    """Convert region geometrical data to geojson format"""

    def make_dict(id_, verts):
        d = dict()
        c = list(map(lambda x: list(x), list(map(lambda v: [int(v[0]), int(v[1])], verts))))
        d["properties"] = {"id": id_}
        d["geometry"] = {"type": "Polygon", "coordinates": c}
        return d

    if use_hull:
        coordinates = r.hull
    else:
        coordinates = r.coordinates
    return [make_dict(id_, verts) for id_, verts in enumerate(coordinates)]



import random
from itertools import product
from typing import Tuple

import numpy as np
from numpy import zeros, array
from numpy.random import rand, normal, poisson
from pandas import DataFrame, concat
from skimage.exposure import rescale_intensity
from skimage.filters import gaussian
from slicedimage import Tile, TileSet

from starfish.constants import Indices, Coordinates
from starfish.image import ImageStack
from starfish.io import Stack


# TODO sofroniewn: doc me!
def synthesize() -> Tuple[Stack, list]:
    """Synthesize synthetic spatial image-based transcriptomics data

    Returns
    -------
    Stack :
        starfish Stack containing synthetic spots
    list :
        codebook matching the synthetic data

    """

    # set random seed so that data is consistent across tests
    random.seed(2)
    np.random.seed(2)

    NUM_HYB = 4
    NUM_CH = 2
    NUM_Z = 1
    HEIGHT = 100
    WIDTH = 100

    assert WIDTH == HEIGHT  # for compatibility with the parameterization of the code

    def choose(n, k):
        if n == k:
            return [[1] * k]
        subsets = [[0] + a for a in choose(n - 1, k)]
        if k > 0:
            subsets += [[1] + a for a in choose(n - 1, k - 1)]
        return subsets

    def graham_sloane_codes(n):
        # n is length of codeword
        # number of on bits is 4
        def code_sum(codeword):
            return sum([i * c for i, c in enumerate(codeword)]) % n
        return [c for c in choose(n, 4) if code_sum(c) == 0]

    p = {
        # number of on bits (not used with current codebook)
        'N_high': 4,
        # length of barcode
        'N_barcode': NUM_CH * NUM_HYB,
        # mean number of flourophores per transcripts - depends on amplification strategy (e.g HCR, bDNA)
        'N_flour': 200,
        # mean number of photons per flourophore - depends on exposure time, bleaching rate of dye
        'N_photons_per_flour': 50,
        # mean number of background photons per pixel - depends on tissue clearing and autoflourescence
        'N_photon_background': 1000,
        # quantum efficiency of the camera detector units number of electrons per photon
        'detection_efficiency': .25,
        # camera read noise per pixel in units electrons
        'N_background_electrons': 1,
        # number of RNA puncta; keep this low to reduce overlap probability
        'N_spots': 20,
        # height and width of image in pixel units
        'N_size': WIDTH,
        # standard devitation of gaussian in pixel units
        'psf': 2,
        # dynamic range of camera sensor 37,000 assuming a 16-bit AD converter
        'graylevel': 37000.0 / 2 ** 16,
        # 16-bit AD converter
        'bits': 16
    }

    codebook = graham_sloane_codes(p['N_barcode'])

    def generate_spot(p):
        position = rand(2)
        gene = random.choice(range(len(codebook)))
        barcode = array(codebook[gene])
        photons = [poisson(p['N_photons_per_flour']) * poisson(p['N_flour']) * b for b in barcode]
        return DataFrame({'position': [position], 'barcode': [barcode], 'photons': [photons], 'gene': gene})

    # right now there is no jitter on x-y positions of the spots, we might want to make it a vector
    spots = concat([generate_spot(p) for _ in range(p['N_spots'])])  # type: ignore

    image = zeros((p['N_barcode'], p['N_size'], p['N_size'],))

    for s in spots.itertuples():
        image[:, int(p['N_size'] * s.position[0]), int(p['N_size'] * s.position[1])] = s.photons

    image_with_background = image + poisson(p['N_photon_background'], size=image.shape)
    filtered = array([gaussian(im, p['psf']) for im in image_with_background])
    filtered = filtered * p['detection_efficiency'] + normal(scale=p['N_background_electrons'], size=filtered.shape)
    signal = np.array([(x / p['graylevel']).astype(int).clip(0, 2 ** p['bits']) for x in filtered])

    def select_uint_dtype(array):
        """choose appropriate dtype based on values of an array"""
        max_val = np.max(array)
        for dtype in [np.uint8, np.uint16, np.uint32, np.uint64]:
            if max_val <= dtype(-1):
                return array.astype(dtype)
        raise ValueError('value exceeds dynamic range of largest numpy type')

    corrected_signal = select_uint_dtype(signal)
    rescaled_signal: np.ndarray = rescale_intensity(corrected_signal)

    # set up the tile set
    image_data = TileSet(
        {Coordinates.X, Coordinates.Y, Indices.HYB, Indices.CH, Indices.Z},
        {
            Indices.HYB: NUM_HYB,
            Indices.CH: NUM_CH,
            Indices.Z: NUM_Z,
        },
        default_tile_shape=(HEIGHT, WIDTH),
    )

    # fill the TileSet
    experiment_indices = list(product(range(NUM_HYB), range(NUM_CH), range(NUM_Z)))
    for i, (hyb, ch, z) in enumerate(experiment_indices):

        tile = Tile(
            {
                Coordinates.X: (0.0, 0.001),
                Coordinates.Y: (0.0, 0.001),
                Coordinates.Z: (0.0, 0.001),
            },
            {
                Indices.HYB: hyb,
                Indices.CH: ch,
                Indices.Z: z,
            }
        )
        tile.numpy_array = rescaled_signal[i]

        image_data.add_tile(tile)

    data_stack = ImageStack(image_data)

    # make a max projection and pretend that's the dots image, which we'll create another ImageStack for this
    dots_data = TileSet(
        {Coordinates.X, Coordinates.Y, Indices.HYB, Indices.CH, Indices.Z},
        {
            Indices.HYB: 1,
            Indices.CH: 1,
            Indices.Z: 1,
        },
        default_tile_shape=(HEIGHT, WIDTH),
    )
    tile = Tile(
        {
            Coordinates.X: (0.0, 0.001),
            Coordinates.Y: (0.0, 0.001),
            Coordinates.Z: (0.0, 0.001),
        },
        {
            Indices.HYB: 0,
            Indices.CH: 0,
            Indices.Z: 0,
        }
    )

    tile.numpy_array = np.max(rescaled_signal, 0)

    dots_data.add_tile(tile)
    dots_stack = ImageStack(dots_data)

    # TODO can we mock up a nuclei image somehow?

    # put the data together into a top-level Stack
    results = Stack.from_data(data_stack, aux_dict={'dots': dots_stack})

    # make the codebook(s)
    codebook = []
    for _, code_record in spots.iterrows():
        codeword = []
        for code_value, (hyb, ch, z) in zip(code_record['barcode'], experiment_indices):
            if code_value != 0:
                codeword.append({
                    Indices.HYB: hyb,
                    Indices.CH: ch,
                    Indices.Z: z,
                    "v": code_value
                })
        codebook.append(
            {
                'codeword': codeword,
                'gene_name': code_record['gene']
            }
        )

    return results, codebook



import pandas as pd


class ValidatedTable:

    required_fields = NotImplemented

    """
    This base class defines common methods for the json outputs of the starfish package, each of which is stored
    internally as a pandas dataframe.

    Methods
    -------
    _validate_table(table: pd.DataFrame, required_fields)
        Each subclass of this base must define required_fields as a class object. When initializing a new object,
        table is checked to verify it contains the required fields
    save(output_file_name: str)
        Save the table to json
    load(json_file: str)
        Load a table from json

    """

    def __init__(self, table: pd.DataFrame, required_fields: set) -> None:
        """

        Parameters
        ----------
        table : pd.DataFrame
            Data to validate
        required_fields : set
            fields table is required to have (class object from implementing subclass)

        """
        self._validate_table(table, required_fields)
        self._data = table

    @property
    def data(self):
        return self._data

    @staticmethod
    def _validate_table(table: pd.DataFrame, required_fields: set):
        missing_fields = required_fields.difference(table.columns)
        if missing_fields:
            raise ValueError(f'input table with columns {table.columns} is missing {missing_fields} required fields')

    def save(self, output_file_name: str) -> None:
        """Save class data to json

        Parameters
        ----------
        output_file_name : str
            Name for output json file

        """
        self.data.to_json(output_file_name, orient='records')

    @classmethod
    def load(cls, json_file: str):
        """Load a ValidatedTable from json

        Parameters
        ----------
        json_file : str
            json file to read

        Returns
        -------
        ValidatedTable :
            Table containing the loaded data

        """

        return cls(pd.read_json(json_file))  # type: ignore # all the subclasses have this signature for constructor.



from starfish.pipeline.algorithmbase import AlgorithmBase


class RegistrationAlgorithmBase(AlgorithmBase):
    def register(self, stack):
        """Performs registration on the stack provided."""
        raise NotImplementedError()



import argparse
import os

from starfish.io import Stack
from starfish.pipeline.features.pixels import _base
from starfish.pipeline.pipelinecomponent import PipelineComponent
from starfish.util.argparse import FsExistsType
from . import pixel_spot_detector


class PixelSpotDetector(PipelineComponent):

    pixel_finder_algorithm_class = None
    pixel_finder_group: argparse.ArgumentParser

    @classmethod
    def implementing_algorithms(cls):
        return _base.PixelFinderAlgorithmBase.__subclasses__()

    @classmethod
    def add_to_parser(cls, subparsers):
        """Adds the pixel finder component to the CLI argument parser."""
        pixel_finder_group = subparsers.add_parser("detect_pixels")
        pixel_finder_group.add_argument("-i", "--input", type=FsExistsType(), required=True)
        pixel_finder_group.add_argument("-o", "--output", required=True)
        pixel_finder_group.set_defaults(starfish_command=PixelSpotDetector._cli)
        pixel_finder_subparsers = pixel_finder_group.add_subparsers(dest="pixel_finder_algorithm_class")

        for algorithm_cls in cls.algorithm_to_class_map().values():
            group_parser = pixel_finder_subparsers.add_parser(algorithm_cls.get_algorithm_name())
            group_parser.set_defaults(pixel_finder_algorithm_class=algorithm_cls)
            algorithm_cls.add_arguments(group_parser)

        cls.pixel_finder_group = pixel_finder_group

    @classmethod
    def _cli(cls, args, print_help=False):
        """Runs the pixel finder component based on parsed arguments."""

        if args.pixel_finder_algorithm_class is None or print_help:
            cls.pixel_finder_group.print_help()
            cls.pixel_finder_group.exit(status=2)

        print('Detecting Pixels...')
        s = Stack()
        s.read(args.input)
        instance = args.pixel_finder_algorithm_class(**vars(args))
        # TODO does this work?
        pixel_attributes, encoded_pixels = instance.find(s)

        # TODO ambrosejcarr: this still needs to be added back.
        # if args.show:
        #     encoded_pixels.show(figsize=(10, 10))

        path = os.path.join(args.output, 'pixels.geojson')
        print(f"Writing | pixels geojson to: {path}")
        pixel_attributes.save_geojson(path)

        path = os.path.join(args.output, 'pixels.json')
        print(f"Writing | spot_id | x | y | z | to: {path}")
        pixel_attributes.save(path)

        path = os.path.join(args.output, 'encoder_table.json')
        print(f"Writing | spot_id | hyb | ch | val | to: {path}")
        encoded_pixels.save(path)



from typing import List

import numpy as np
from trackpy import locate

from starfish.image import ImageStack
from starfish.pipeline.features.spot_attributes import SpotAttributes
from ._base import SpotFinderAlgorithmBase


class LocalMaxPeakFinder(SpotFinderAlgorithmBase):

    def __init__(
            self, spot_diameter, min_mass, max_size, separation, percentile=0, noise_size=None, smoothing_size=None,
            threshold=None, preprocess: bool=False, is_volume: bool=False, verbose=False, **kwargs
    ) -> None:
        """Local max peak finding algorithm

        This is a wrapper for `trackpy.locate`

        Parameters
        ----------

        spot_diameter : odd integer or tuple of odd integers.
            This may be a single number or a tuple giving the feature’s extent in each dimension, useful when the
            dimensions do not have equal resolution (e.g. confocal microscopy). The tuple order is the same as the
            image shape, conventionally (z, y, x) or (y, x). The number(s) must be odd integers. When in doubt, round
            up.
        min_mass : float, optional
            The minimum integrated brightness. This is a crucial parameter for eliminating spurious features.
            Recommended minimum values are 100 for integer images and 1 for float images. Defaults to 0 (no filtering).
        max_size : float
            maximum radius-of-gyration of brightness, default None
        separation : float or tuple
            Minimum separtion between features. Default is diameter + 1. May be a tuple, see diameter for details.
        percentile : float
            Features must have a peak brighter than pixels in this percentile. This helps eliminate spurious peaks.
        noise_size : float or tuple
            Width of Gaussian blurring kernel, in pixels Default is 1. May be a tuple, see diameter for details.
        smoothing_size : float or tuple
            The size of the sides of the square kernel used in boxcar (rolling average) smoothing, in pixels Default
            is diameter. May be a tuple, making the kernel rectangular.
        threshold : float
            Clip bandpass result below this value. Thresholding is done on the already background-subtracted image.
            By default, 1 for integer images and 1/255 for float images.
        preprocess : boolean
            Set to False to turn off bandpass preprocessing.
        max_iterations : integer
            max number of loops to refine the center of mass, default 10
        is_volume : bool
            if True, run the algorithm on 3d volumes of the provided stack
        verbose : bool
            If True, report the percentage completed (default = False) during processing


        See Also
        --------
        trackpy locate: http://soft-matter.github.io/trackpy/dev/generated/trackpy.locate.html

        """

        self.diameter = spot_diameter
        self.minmass = min_mass
        self.maxsize = max_size
        self.separation = separation
        self.noise_size = noise_size
        self.smoothing_size = smoothing_size
        self.percentile = percentile
        self.threshold = threshold
        self.preprocess = preprocess
        self.is_volume = is_volume
        self.verbose = verbose

    # # TODO ambrosejcarr: make this generalize to smFISH methods
    # def encode(self, spot_attributes: SpotAttributes):
    #     spot_table = spot_attributes.data
    #     spot_table['barcode_index'] = np.ones(spot_table.shape[0])

    def find_attributes(self, image: np.ndarray) -> SpotAttributes:
        """

        Parameters
        ----------
        image : np.ndarray
            two- or three-dimensional numpy array containing spots to detect

        Returns
        -------
        SpotAttributes :
            spot attributes table for all detected spots

        """
        attributes = locate(
            image,
            diameter=self.diameter,
            minmass=self.minmass,
            maxsize=self.maxsize,
            separation=self.separation,
            noise_size=self.noise_size,
            smoothing_size=self.smoothing_size,
            threshold=self.threshold,
            percentile=self.percentile,
            preprocess=self.preprocess
        )

        new_colnames = ['x', 'y', 'intensity', 'r', 'eccentricity', 'signal', 'raw_mass', 'ep']
        if len(image.shape) == 3:
            attributes.columns = ['z'] + new_colnames
        else:
            attributes.columns = new_colnames

        attributes['spot_id'] = np.arange(attributes.shape[0])
        return SpotAttributes(attributes)

    def find(self, stack: ImageStack):
        """
        Find spots.

        Parameters
        ----------
        stack : ImageStack
            Stack where we find the spots in.
        """
        spot_attributes: List[SpotAttributes] = stack.apply(
            self.find_attributes, in_place=False, is_volume=self.is_volume, verbose=self.verbose)

        # TODO ambrosejcarr: do we need to find spots in the aux_dict too?

        # TODO ambrosejcarr: this is where development stopped; spot_attributes is correct, but translating
        # spot_attributes into an encoder_table is tricky without first implementing the new codebook. Do that first.
        # create an encoded table
        # encoded_spots = self.encode(spot_attributes.data)

        return spot_attributes

    @classmethod
    def add_arguments(cls, group_parser):
        group_parser.add_argument("--spot-diameter", type=str, help='expected spot size')
        group_parser.add_argument("--min-mass", default=4, type=int, help="minimum integrated spot intensity")
        group_parser.add_argument("--max-size", default=6, type=int, help="maximum radius of gyration of brightness")
        group_parser.add_argument("--separation", default=5, type=float, help="minimum distance between spots")
        group_parser.add_argument(
            "--noise-size", default=None, type=int, help="width of gaussian blurring kernel, in pixels")
        group_parser.add_argument(
            "--smoothing-size", default=None, type=int,
            help="odd integer. Size of boxcar (moving average) filter in pixels. Default is the Diameter")
        group_parser.add_argument(
            "--preprocess", action="store_true", help="if passed, gaussian and boxcar filtering are applied")
        group_parser.add_argument(
            "--show", default=False, action='store_true', help="display results visually")
        group_parser.add_argument(
            "--percentile", default=None, type=float,
            help="clip bandpass below this value. Thresholding is done on already background-subtracted images. "
                 "default 1 for integer images and 1/255 for float"
        )



from starfish.pipeline.features.spots.detector.gaussian import GaussianSpotDetector
from starfish.pipeline.filter.white_tophat import WhiteTophat
from starfish.pipeline.registration.fourier_shift import FourierShiftRegistration
from starfish.test.dataset_fixtures import labeled_synthetic_dataset


def test_merfish_pipeline():
    stack = labeled_synthetic_dataset()

    fsr = FourierShiftRegistration(upsampling=1000, reference_stack=stack.auxiliary_images['dots'])
    fsr.register(stack.image)

    wth = WhiteTophat(disk_size=15)
    wth.filter(stack.image)
    for image in stack.auxiliary_images.values():
        wth.filter(image)

    gsd = GaussianSpotDetector(
        blobs_stack=stack.auxiliary_images['dots'], min_sigma=2, max_sigma=10, num_sigma=10, threshold=0.1)
    spot_attributes, encoded_spots = gsd.find(stack.image)

    assert spot_attributes.data.shape[0] == 19



import numpy as np

from starfish.image import ImageStack
from ._base import FilterAlgorithmBase


class WhiteTophat(FilterAlgorithmBase):
    """
    Performs "white top hat" filtering of an image to enhance spots. "White top hat filtering" finds spots that are both
    smaller and brighter than their surroundings.

    See Also
    --------
    https://en.wikipedia.org/wiki/Top-hat_transform
    """

    def __init__(self, disk_size, **kwargs):
        """Instance of a white top hat morphological masking filter which masks objects larger than `disk_size`

        Parameters
        ----------
        disk_size : int
            diameter of the morphological masking disk in pixels

        """
        self.disk_size = disk_size

    @classmethod
    def add_arguments(cls, group_parser):
        group_parser.add_argument(
            "--disk-size", default=15, type=int, help="diameter of morphological masking disk in pixels")

    def filter(self, stack: ImageStack) -> None:
        """Perform in-place filtering of an image stack and all contained aux images

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered

        """
        from scipy.ndimage.filters import maximum_filter, minimum_filter
        from skimage.morphology import disk

        def white_tophat(image):
            if image.dtype.kind != "u":
                raise TypeError("images should be stored in an unsigned integer array")
            structuring_element = disk(self.disk_size)
            min_filtered = minimum_filter(image, footprint=structuring_element)
            max_filtered = maximum_filter(min_filtered, footprint=structuring_element)
            filtered_image = image - np.minimum(image, max_filtered)
            return filtered_image

        stack.apply(white_tophat)



from enum import Enum


class Coordinates(Enum):
    X = 'x'
    Y = 'y'
    Z = 'z'

    def __hash__(self):
        return self.value.__hash__()

    def __eq__(self, other):
        if isinstance(other, Coordinates) or isinstance(other, str):
            return self.value == other
        return False


class Indices(Enum):
    HYB = 'h'
    CH = 'c'
    Z = 'z'

    def __hash__(self):
        return self.value.__hash__()

    def __eq__(self, other):
        if isinstance(other, Indices) or isinstance(other, str):
            return self.value == other
        return False



import collections
import os
from functools import partial
from itertools import product
from typing import Any, Callable, Iterable, Iterator, Mapping, MutableSequence, Optional, Sequence, Tuple, Union
from warnings import warn

import numpy
import pandas as pd
from scipy.stats import scoreatpercentile
from skimage import exposure
from slicedimage import Reader, Writer
from slicedimage.io import resolve_path_or_url
from tqdm import tqdm

from starfish.constants import Coordinates, Indices
from starfish.errors import DataFormatWarning
from starfish.pipeline.features.spot_attributes import SpotAttributes


class ImageStack:
    """Container for a TileSet (field of view)

    Methods
    -------
    get_slice    retrieve a slice of the image tensor
    set_slice    set a slice of the image tensor
    apply        apply a 2d or 3d function across all Tiles in the image tensor
    max_proj     return a max projection over one or more axis of the image tensor
    show_stack   show an interactive, pageable view of the image tensor, or a slice of the image tensor
    write        save the (potentially modified) image tensor to disk

    Properties
    ----------
    num_chs      the number of channels stored in the image tensor
    num_hybs     the number of hybridization rounds stored in the image tensor
    num_zlayers  the number of z-layers stored in the image tensor
    numpy_array  the 5-d image tensor is stored in this array
    raw_shape    the shape of the image tensor (in integers)
    shape        the shape of the image tensor by categorical index (channels, hybridization rounds, z-layers)
    """

    AXES_MAP = {
        Indices.HYB: 0,
        Indices.CH: 1,
        Indices.Z: 2,
    }
    N_AXES = max(AXES_MAP.values()) + 1

    def __init__(self, image_partition):
        self._image_partition = image_partition
        self._num_hybs = image_partition.get_dimension_shape(Indices.HYB)
        self._num_chs = image_partition.get_dimension_shape(Indices.CH)
        if Indices.Z in image_partition.dimensions:
            self._num_zlayers = image_partition.get_dimension_shape(Indices.Z)
        else:
            self._num_zlayers = 1
        self._tile_shape = image_partition.default_tile_shape

        # Examine the tiles to figure out the right kind (int, float, etc.) and size.  We require that all the tiles
        # have the same kind of data type, but we do not require that they all have the same size of data type.  The
        # allocated array is the highest size we encounter.
        kind = None
        max_size = 0
        for tile in self._image_partition.tiles():
            dtype = tile.numpy_array.dtype
            if kind is None:
                kind = dtype.kind
            else:
                if kind != dtype.kind:
                    raise TypeError("All tiles should have the same kind of dtype")
            if dtype.itemsize > max_size:
                max_size = dtype.itemsize
            if self._tile_shape is None:
                self._tile_shape = tile.tile_shape
            elif tile.tile_shape is not None and self._tile_shape != tile.tile_shape:
                raise ValueError("Starfish does not support tiles that are not identical in shape")

        # now that we know the tile data type (kind and size), we can allocate the data array.
        self._data = numpy.zeros(
            shape=(self._num_hybs, self._num_chs, self._num_zlayers) + self._tile_shape,
            dtype=numpy.dtype(f"{kind}{max_size}")
        )

        # iterate through the tiles and set the data.
        for tile in self._image_partition.tiles():
            h = tile.indices[Indices.HYB]
            c = tile.indices[Indices.CH]
            zlayer = tile.indices.get(Indices.Z, 0)
            data = tile.numpy_array
            if max_size != data.dtype.itemsize:
                if data.dtype.kind == "i" or data.dtype.kind == "u":
                    # fixed point can be done with a simple multiply.
                    src_range = numpy.iinfo(data.dtype).max - numpy.iinfo(data.dtype).min + 1
                    dst_range = numpy.iinfo(self._data.dtype).max - numpy.iinfo(self._data.dtype).min + 1
                    data = data * (dst_range / src_range)
                warn(
                    f"Tile "
                    f"(H: {tile.indices[Indices.HYB]} C: {tile.indices[Indices.CH]} Z: {tile.indices[Indices.Z]}) has "
                    f"dtype {data.dtype}.  One or more tiles is of a larger dtype {self._data.dtype}.",
                    DataFormatWarning)
            self.set_slice(indices={Indices.HYB: h, Indices.CH: c, Indices.Z: zlayer}, data=data)
        # set_slice will mark the data as needing writeback, so we need to unset that.
        self._data_needs_writeback = False

    @classmethod
    def from_url(cls, url: str, baseurl: Optional[str]):
        """
        Constructs an ImageStack object from a URL and a base URL.

        The following examples will all load from the same location:
          1. url: https://www.example.com/images/hybridization.json  baseurl: None
          2. url: https://www.example.com/images/hybridization.json  baseurl: I_am_ignored
          3. url: hybridization.json  baseurl: https://www.example.com/images
          4. url: images/hybridization.json  baseurl: https://www.example.com

        Parameters:
        -----------
        url : str
            Either an absolute URL or a relative URL referring to the image to be read.
        baseurl : Optional[str]
            If url is a relative URL, then this must be provided.  If url is an absolute URL, then this parameter is
            ignored.
        """
        image_partition = Reader.parse_doc(url, baseurl)

        return cls(image_partition)

    @classmethod
    def from_path_or_url(cls, url_or_path: str) -> "ImageStack":
        """
        Constructs an ImageStack object from an absolute URL or a filesystem path.

        The following examples will all load from the same location:
          1. url_or_path: file:///Users/starfish-user/images/hybridization.json
          2. url_or_path: /Users/starfish-user/images/hybridization.json

        Parameters:
        -----------
        url_or_path : str
            Either an absolute URL or a filesystem path to an imagestack.
        """
        _, relativeurl, baseurl = resolve_path_or_url(url_or_path)
        return cls.from_url(relativeurl, baseurl)

    @property
    def numpy_array(self):
        """Retrieves a view of the image data as a numpy array."""
        result = self._data.view()
        result.setflags(write=False)
        return result

    @numpy_array.setter
    def numpy_array(self, data):
        """Sets the image's data from a numpy array.  The numpy array is advised to be immutable afterwards."""
        self._data = data.view()
        self._data_needs_writeback = True
        data.setflags(write=False)

    def get_slice(
            self,
            indices: Mapping[Indices, Union[int, slice]]
    ) -> Tuple[numpy.ndarray, Sequence[Indices]]:
        """
        Given a dictionary mapping the index name to either a value or a slice range, return a numpy array representing
        the slice, and a list of the remaining axes beyond the normal x-y tile.

        Example:
            ImageStack axes: H, C, and Z with shape 3, 4, 5, respectively.
            ImageStack Implicit axes: X, Y with shape 10, 20, respectively.
            Called to slice with indices {Z: 5}.
            Result: a 4-dimensional numpy array with shape (3, 4, 20, 10) and the remaining axes [H, C].

        Example:
            Original axes: H, C, and Z.
            Implicit axes: X, Y.
            Called to slice with indices {Z: 5, C: slice(2, 4)}.
            Result: a 4-dimensional numpy array with shape (3, 2, 20, 10) and the remaining axes [H, C].
        """
        slice_list, axes = self._build_slice_list(indices)
        result = self._data[slice_list]
        result.setflags(write=False)
        return result, axes

    def set_slice(
            self,
            indices: Mapping[Indices, Union[int, slice]],
            data: numpy.ndarray,
            axes: Sequence[Indices]=None):
        """
        Given a dictionary mapping the index name to either a value or a slice range and a source numpy array, set the
        slice of the array of this ImageStack to the values in the source numpy array.  If the optional parameter axes
        is provided, that represents the axes of the numpy array beyond the x-y tile.

        Example:
            ImageStack axes: H, C, and Z with shape 3, 4, 5, respectively.
            ImageStack Implicit axes: X, Y with shape 10, 20, respectively.
            Called to set a slice with indices {Z: 5}.
            Data: a 4-dimensional numpy array with shape (3, 4, 10, 20)
            Result: Replace the data for Z=5.

        Example:
            ImageStack axes: H, C, and Z. (shape 3, 4, 5)
            ImageStack Implicit axes: X, Y. (shape 10, 20)
            Called to set a slice with indices {Z: 5, C: slice(2, 4)}.
            Data: a 4-dimensional numpy array with shape (3, 2, 10, 20)
            Result: Replace the data for Z=5, C=2-3.
        """
        slice_list, expected_axes = self._build_slice_list(indices)

        if axes is not None:
            if len(axes) != len(data.shape) - 2:
                raise ValueError("data shape ({}) should be the axes ({}) and (x,y).".format(data.shape, axes))
            move_src = list()
            move_dst = list()
            for src_idx, axis in enumerate(axes):
                try:
                    dst_idx = expected_axes.index(axis)
                except ValueError:
                    raise ValueError("Unexpected axis {}.  Expecting only {}.".format(axis, expected_axes))
                if src_idx != dst_idx:
                    move_src.append(src_idx)
                    move_dst.append(dst_idx)

            if len(move_src) != 0:
                data = data.view()
                numpy.moveaxis(data, move_src, move_dst)

        if self._data[slice_list].shape != data.shape:
            raise ValueError("source shape {} mismatches destination shape {}".format(
                data.shape, self._data[slice_list].shape))

        self._data[slice_list] = data
        self._data_needs_writeback = True

    def show_stack(
            self, indices: Mapping[Indices, Union[int, slice]],
            color_map: str= 'gray', figure_size: Tuple[int, int]=(10, 10),
            show_spots: Optional[SpotAttributes]=None,
            rescale: bool=False, p_min: Optional[float]=None, p_max: Optional[float]=None, **kwargs):
        """Create an interactive visualization of an image stack

        Produces a slider that flips through the selected volume tile-by-tile. Supports manual adjustment of dynamic
        range.

        Parameters
        ----------
        indices : Mapping[Indices, Union[int, slice]],
            Indices to select a volume to visualize. Passed to `Image.get_slice()`.
            See `Image.get_slice()` for examples.
        color_map : str (default = 'gray')
            string id of a matplotlib colormap
        figure_size : Tuple[int, int] (default = (10, 10))
            size of the figure in inches
        show_spots : Optional[SpotAttributes]
            [Preliminary functionality] if provided, should be a SpotAttribute table that corresponds
            to the volume being displayed. This will be paired automatically in the future.
        rescale : bool (default = False)
            if True, rescale the data to exclude high and low-value outliers (see skimage.exposure.rescale_intensity).
        p_min: float
            clip values below this intensity percentile. If provided, overrides rescale, above. (default = None)
        p_max: float
            clip values above this intensity percentile. If provided, overrides rescale, above. (default = None)

        Raises
        ------
        ValueError :
            User must select one of rescale or p_min/p_max to adjust the image dynamic range. If both are selected, a
            ValueError is raised.

        Notes
        -----
        For this widget to function interactively in the notebook, after ipywidgets has been installed, the user must
        register the widget with jupyter by typing the following command into the terminal:
        jupyter nbextension enable --py widgetsnbextension

        """

        from ipywidgets import interact
        import matplotlib.pyplot as plt

        if not indices:
            raise ValueError('indices may not be an empty dict or None')

        # get the requested chunk, linearize the remaining data into a sequence of tiles
        data, remaining_inds = self.get_slice(indices)

        # identify the dimensionality of data with all dimensions other than x, y linearized
        n = numpy.dot(*data.shape[:-2])

        # linearize the array
        linear_view: numpy.ndarray = data.reshape((n,) + data.shape[-2:])

        # set the labels for the linearized tiles

        labels = []
        for index, size in zip(remaining_inds, data.shape[:-2]):
            labels.append([f'{index}{n}' for n in range(size)])
        labels = list(product(*labels))

        n = linear_view.shape[0]

        if rescale and any((p_min, p_max)):
            raise ValueError('select one of rescale and p_min/p_max to rescale image, not both.')

        elif rescale is not None:
            print("Rescaling ...")
            vmin, vmax = scoreatpercentile(data, (0.5, 99.5))
            linear_view = exposure.rescale_intensity(
                linear_view,
                in_range=(vmin, vmax),
                out_range=numpy.float32
            ).astype(numpy.float32)

        elif p_min or p_max:
            print("Clipping ...")
            a_min, a_max = scoreatpercentile(
                linear_view,
                (p_min if p_min else 0, p_max if p_max else 100)
            )
            linear_view = numpy.clip(linear_view, a_min=a_min, a_max=a_max)

        show_spot_function = self._show_spots

        def show_plane(ax, plane, plane_index, cmap="gray", title=None):
            ax.imshow(plane, cmap=cmap)
            if show_spots:
                # this is slow. This link might have something to help:
                # https://bastibe.de/2013-05-30-speeding-up-matplotlib.html
                show_spot_function(show_spots.data, ax=ax, z=plane_index, **kwargs)
            ax.set_xticks([])
            ax.set_yticks([])

            if title:
                ax.set_title(title)

        @interact(plane_index=(0, n - 1))
        def display_slice(plane_index=0):
            fig, ax = plt.subplots(figsize=figure_size)
            show_plane(ax, linear_view[plane_index], plane_index, title=f'{labels[plane_index]}', cmap=color_map)
            plt.show()

        return display_slice

    @staticmethod
    def _show_spots(result_df, ax, z=None, size=1, z_dist=1.5, scale_radius=5) -> None:
        """function to plot spot finding results on top of any image as hollow red circles

        called spots are colored by category

        Parameters:
        -----------
        img : np.ndarray[Any]
            2-d image of any dtype
        result_df : pd.Dataframe
            result dataframe containing spot calls that correspond to the image channel
        z : Optional[int]
            If provided, z-plane to plot spot calls for. Default (None): plot all provided spots.
        size : int
            width of line to plot around the identified spot
        z_dist : float
            plot spots if within this distance of the z-plane. Ignored if z is not passed.
        vmin, vmax : int
            clipping thresholds for the image plot
        ax, matplotlib.Axes.Axis
            axis to plot spots on

        """
        import matplotlib.pyplot as plt

        if z is not None and z in result_df.columns:
            inds = numpy.abs(result_df['z'] - z) < z_dist
        else:
            inds = numpy.ones(result_df.shape[0]).astype(bool)

        # get the data needed to plot
        selected = result_df.loc[inds, ['r', 'x', 'y']]

        for i in numpy.arange(selected.shape[0]):
            r, x, y = selected.iloc[i, :]  # radius is a duplicate, and is present twice
            c = plt.Circle((x, y), r * scale_radius, color='r', linewidth=size, fill=False)
            ax.add_patch(c)

    def _build_slice_list(
            self,
            indices: Mapping[Indices, Union[int, slice]]
    ) -> Tuple[Tuple[Union[int, slice], ...], Sequence[Indices]]:
        slice_list: MutableSequence[Union[int, slice]] = [
            slice(None, None, None)
            for _ in range(ImageStack.N_AXES)
        ]
        axes = []
        removed_axes = set()
        for name, value in indices.items():
            idx = ImageStack.AXES_MAP[name]
            if not isinstance(value, slice):
                removed_axes.add(name)
            slice_list[idx] = value

        for dimension_value, dimension_name in sorted([
            (dimension_value, dimension_name)
            for dimension_name, dimension_value in ImageStack.AXES_MAP.items()
        ]):
            if dimension_name not in removed_axes:
                axes.append(dimension_name)

        return tuple(slice_list), axes

    def _iter_indices(self, is_volume: bool=False) -> Iterator[Mapping[Indices, int]]:
        """Iterate over indices of image tiles or image volumes if is_volume is True

        Parameters
        ----------
        is_volume, bool
            If True, yield indices necessary to extract volumes from self, else return
            indices for tiles

        Yields
        ------
        Dict[str, int]
            Mapping of dimension name to index

        """
        for hyb in numpy.arange(self.shape[Indices.HYB]):
            for ch in numpy.arange(self.shape[Indices.CH]):
                if is_volume:
                    yield {Indices.HYB: hyb, Indices.CH: ch}
                else:
                    for z in numpy.arange(self.shape[Indices.Z]):
                        yield {Indices.HYB: hyb, Indices.CH: ch, Indices.Z: z}

    def _iter_tiles(
            self, indices: Iterable[Mapping[Indices, Union[int, slice]]]
    ) -> Iterable[numpy.ndarray]:
        """Given an iterable of indices, return a generator of numpy arrays from self

        Parameters
        ----------
        indices, Iterable[Mapping[str, int]]
            Iterable of indices that map a dimension (str) to a value (int)

        Yields
        ------
        numpy.ndarray
            Numpy array that corresponds to provided indices
        """
        for inds in indices:
            array, axes = self.get_slice(inds)
            yield array

    def apply(self, func, is_volume=False, in_place=True, verbose: bool=False, **kwargs):
        """Apply func over all tiles or volumes in self

        Parameters
        ----------
        func : Callable
            Function to apply. must expect a first argument which is a 2d or 3d numpy array (see is_volume) and return a
            numpy.ndarray. If inplace is True, must return an array of the same shape.
        is_volume : bool
            (default False) If True, pass 3d volumes (x, y, z) to func
        in_place : bool
            (default True) If True, function is executed in place. If n_proc is not 1, the tile or
            volume will be copied once during execution. If false, the outputs of the function executed on individual
            tiles or volumes will be output as a list
        verbose : bool
            If True, report on the percentage completed (default = False) during processing
        kwargs : dict
            Additional arguments to pass to func

        Returns
        -------
        Optional[List[Tuple[np.ndarray, Mapping[Indices: Union[int, slice]]]]
            If inplace is False, return the results of applying func to stored image data
        """
        mapfunc: Callable = map  # TODO: ambrosejcarr posix-compliant multiprocessing
        indices = list(self._iter_indices(is_volume=is_volume))

        if verbose:
            tiles = tqdm(self._iter_tiles(indices))
        else:
            tiles = self._iter_tiles(indices)

        applyfunc: Callable = partial(func, **kwargs)
        results = mapfunc(applyfunc, tiles)

        # TODO ttung: this should return an ImageStack, not a bunch of indices.
        if not in_place:
            return list(zip(results, indices))

        for r, inds in zip(results, indices):
            self.set_slice(inds, r)

    @property
    def tile_metadata(self) -> pd.DataFrame:
        """return a table containing Tile metadata

        Returns
        -------
        pd.DataFrame :
            dataframe containing per-tile metadata information for each image. Guaranteed to include information on
            channel, hybridization round, z_layer, and barcode index. Also contains any information stored in the
            extras field for each tile in hybridization.json

        """

        data: collections.defaultdict = collections.defaultdict(list)
        index_keys = set(
            key
            for tile in self._image_partition.tiles()
            for key in tile.indices.keys())
        extras_keys = set(
            key
            for tile in self._image_partition.tiles()
            for key in tile.extras.keys())
        duplicate_keys = index_keys.intersection(extras_keys)
        if len(duplicate_keys) > 0:
            duplicate_keys_str = ", ".join([str(key) for key in duplicate_keys])
            raise ValueError(
                f"keys ({duplicate_keys_str}) was found in both the Tile specification and extras field. Tile "
                f"specification keys may not be duplicated in the extras field.")

        for tile in self._image_partition.tiles():
            for k in index_keys:
                data[k].append(tile.indices.get(k, None))
            for k in extras_keys:
                data[k].append(tile.extras.get(k, None))

            if 'barcode_index' not in tile.extras:
                hyb = tile.indices[Indices.HYB]
                ch = tile.indices[Indices.CH]
                z = tile.indices.get(Indices.Z, 0)
                barcode_index = (((z * self.num_hybs) + hyb) * self.num_chs) + ch

                data['barcode_index'].append(barcode_index)

        return pd.DataFrame(data)

    @property
    def raw_shape(self) -> Tuple[int]:
        """
        Returns the shape of the space that this image inhabits.  It does not include the dimensions of the image
        itself.  For instance, if this is an X-Y image in a C-H-Y-X space, then the shape would include the dimensions C
        and H.

        Returns
        -------
        Tuple[int] :
            The sizes of the indices.
        """
        return self._data.shape

    @property
    def shape(self) -> collections.OrderedDict:
        """
        Returns the shape of the space that this image inhabits.  It does not include the dimensions of the image
        itself.  For instance, if this is an X-Y image in a C-H-Y-X space, then the shape would include the dimensions C
        and H.

        Returns
        -------
        An ordered mapping between index names to the size of the index.
        """
        # TODO: (ttung) Note that the return type should be ..OrderedDict[Any, str], but python3.6 has a bug where this
        # breaks horribly.  Can't find a bug id to link to, but see
        # https://stackoverflow.com/questions/41207128/how-do-i-specify-ordereddict-k-v-types-for-mypy-type-annotation
        result: collections.OrderedDict[Any, str] = collections.OrderedDict()
        for name, idx in ImageStack.AXES_MAP.items():
            result[name] = self._data.shape[idx]
        result['y'] = self._data.shape[-2]
        result['x'] = self._data.shape[-1]

        return result

    @property
    def num_hybs(self):
        return self._num_hybs

    @property
    def num_chs(self):
        return self._num_chs

    @property
    def num_zlayers(self):
        return self._num_zlayers

    @property
    def tile_shape(self):
        return self._tile_shape

    def write(self, filepath: str, tile_opener=None) -> None:
        """write the image tensor to disk

        Parameters
        ----------
        filepath : str
            path + prefix for writing the image tensor
        tile_opener : TODO ttung: doc me.

        """
        if self._data_needs_writeback:
            for tile in self._image_partition.tiles():
                h = tile.indices[Indices.HYB]
                c = tile.indices[Indices.CH]
                zlayer = tile.indices.get(Indices.Z, 0)
                tile.numpy_array, axes = self.get_slice(indices={Indices.HYB: h, Indices.CH: c, Indices.Z: zlayer})
                assert len(axes) == 0
            self._data_needs_writeback = False

        seen_x_coords, seen_y_coords, seen_z_coords = set(), set(), set()
        for tile in self._image_partition.tiles():
            seen_x_coords.add(tile.coordinates[Coordinates.X])
            seen_y_coords.add(tile.coordinates[Coordinates.Y])
            z_coords = tile.coordinates.get(Coordinates.Z, None)
            if z_coords is not None:
                seen_z_coords.add(z_coords)

        sorted_x_coords = sorted(seen_x_coords)
        sorted_y_coords = sorted(seen_y_coords)
        sorted_z_coords = sorted(seen_z_coords)
        x_coords_to_idx = {coords: idx for idx, coords in enumerate(sorted_x_coords)}
        y_coords_to_idx = {coords: idx for idx, coords in enumerate(sorted_y_coords)}
        z_coords_to_idx = {coords: idx for idx, coords in enumerate(sorted_z_coords)}

        if tile_opener is None:
            def tile_opener(tileset_path, tile, ext):
                tile_basename = os.path.splitext(tileset_path)[0]
                xcoord = tile.coordinates[Coordinates.X]
                ycoord = tile.coordinates[Coordinates.Y]
                zcoord = tile.coordinates.get(Coordinates.Z, None)
                xcoord = tuple(xcoord) if isinstance(xcoord, list) else xcoord
                ycoord = tuple(ycoord) if isinstance(ycoord, list) else ycoord
                xval = x_coords_to_idx[xcoord]
                yval = y_coords_to_idx[ycoord]
                if zcoord is not None:
                    zval = z_coords_to_idx[zcoord]
                    zstr = "-Z{}".format(zval)
                else:
                    zstr = ""
                return open(
                    "{}-X{}-Y{}{}-H{}-C{}.{}".format(
                        tile_basename,
                        xval,
                        yval,
                        zstr,
                        tile.indices[Indices.HYB],
                        tile.indices[Indices.CH],
                        ext,
                    ),
                    "wb")

        Writer.write_to_path(
            self._image_partition,
            filepath,
            pretty=True,
            tile_opener=tile_opener)

    def max_proj(self, *dims: Indices) -> numpy.ndarray:
        """return a max projection over one or more axis of the image tensor

        Parameters
        ----------
        dims : Indices
            one or more axes to project over

        Returns
        -------
        numpy.ndarray :
            max projection

        """
        axes = list()
        for dim in dims:
            try:
                axes.append(ImageStack.AXES_MAP[dim])
            except KeyError:
                raise ValueError(
                    "Dimension: {} not supported. Expecting one of: {}".format(dim, ImageStack.AXES_MAP.keys()))

        return numpy.max(self._data, axis=tuple(axes))

    def squeeze(self) -> numpy.ndarray:
        """return an array that is linear over categorical dimensions and z

        Returns
        -------
        np.ndarray :
            array of shape (num_hybs + num_channels + num_z_layers, x, y).

        """
        first_dim = self.num_hybs * self.num_chs * self.num_zlayers
        new_shape = (first_dim,) + self.tile_shape
        new_data = self.numpy_array.reshape(new_shape)

        return new_data

    def un_squeeze(self, stack):
        if type(stack) is list:
            stack = numpy.array(stack)

        new_shape = (self.num_hybs, self.num_chs, self.num_zlayers) + self.tile_shape
        res = stack.reshape(new_shape)
        return res



import argparse
from functools import partial
from typing import Callable

import numpy as np

from starfish.image import ImageStack
from starfish.pipeline.filter.gaussian_low_pass import GaussianLowPass
from ._base import FilterAlgorithmBase


class GaussianHighPass(FilterAlgorithmBase):

    def __init__(self, sigma, **kwargs) -> None:
        """Gaussian high pass filter

        Parameters
        ----------
        sigma : int (default = 1)
            standard deviation of gaussian kernel

        """
        self.sigma = sigma

    @classmethod
    def add_arguments(cls, group_parser: argparse.ArgumentParser) -> None:
        group_parser.add_argument(
            "--sigma", default=1, type=int, help="standard deviation of gaussian kernel")

    @staticmethod
    def gaussian_high_pass(img: np.ndarray, sigma) -> np.ndarray:
        """
        Applies a gaussian high pass filter to an image

        Parameters
        ----------
        img : numpy.ndarray
            Image to filter
        sigma : Union[float, int]
            Standard deviation of gaussian kernel

        Returns
        -------
        numpy.ndarray :
            Filtered image, same shape as input

        """
        blurred: np.ndarray = GaussianLowPass.low_pass(img, sigma)

        over_flow_ind: np.ndarray[bool] = img < blurred
        res: np.ndarray = img - blurred
        res[over_flow_ind] = 0

        return res

    def filter(self, stack: ImageStack) -> None:
        """
        Perform in-place filtering of an image stack and all contained aux images.

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.

        """

        high_pass: Callable = partial(self.gaussian_high_pass, sigma=self.sigma)
        stack.apply(high_pass)



from starfish.pipeline.algorithmbase import AlgorithmBase


class SegmentationAlgorithmBase(AlgorithmBase):
    def segment(self, hybridization_stack, nuclei_stack):
        """Performs registration on the stack provided."""
        raise NotImplementedError()



import numpy as np

from starfish.pipeline.features.spots.detector.gaussian import GaussianSpotDetector
from starfish.util.synthesize import SyntheticData


def test_round_trip_synthetic_data():
    sd = SyntheticData(
        n_ch=2,
        n_hyb=3,
        n_spots=1,
        n_codes=4,
        n_photons_background=0,
        background_electrons=0,
        camera_detection_efficiency=1.0,
        gray_level=1,
        ad_conversion_bits=16,
        point_spread_function=(2, 2, 2),
    )

    codebook = sd.codebook()
    intensities = sd.intensities(codebook=codebook)
    spots = sd.spots(intensities=intensities)
    gsd = GaussianSpotDetector(
        min_sigma=1, max_sigma=4, num_sigma=5, threshold=0, blobs_stack=spots)
    calculated_intensities = gsd.find(spots)
    codebook.decode_euclidean(calculated_intensities)

    # applying the gaussian blur to the intensities causes them to be reduced in magnitude, so
    # they won't be the same size, but they should be in the same place, and decode the same
    # way
    spot1, ch1, hyb1 = np.where(intensities.values)
    spot2, ch2, hyb2 = np.where(calculated_intensities.values)
    assert np.array_equal(spot1, spot2)
    assert np.array_equal(ch1, ch2)
    assert np.array_equal(hyb1, hyb2)
    assert np.array_equal(
        intensities.coords[intensities.Constants.GENE],
        calculated_intensities.coords[intensities.Constants.GENE]
    )


# @pytest.mark.skip('long-running integration test, for debugging only')
def test_medium_synthetic_stack():
    n_z = 40
    height = 300
    width = 400
    sigma = 2

    sd = SyntheticData(
        n_hyb=4,
        n_ch=4,
        n_z=n_z,
        height=height,
        width=width,
        n_spots=100,
        n_codes=10,
        point_spread_function=(sigma, sigma, sigma),
    )

    codebook = sd.codebook()
    intensities = sd.intensities(codebook=codebook)

    # some spots won't be detected properly because they will spread outside the image when blurred,
    # so we'll remove those from intensities before we generate spots.

    spot_radius = sigma * np.sqrt(2)  # this is the radius of the spot in pixels

    valid_z = np.logical_and(intensities.z.values > spot_radius,
                             intensities.z.values < (n_z - spot_radius))
    valid_y = np.logical_and(intensities.y.values > spot_radius,
                             intensities.y.values < (height - spot_radius))
    valid_x = np.logical_and(intensities.x.values > spot_radius,
                             intensities.x.values < (width - spot_radius))

    valid_locations = valid_z & valid_y & valid_x
    intensities = intensities[np.where(valid_locations)]
    spots = sd.spots(intensities=intensities)

    gsd = GaussianSpotDetector(min_sigma=1, max_sigma=4, num_sigma=5, threshold=1e-4, blobs_stack=spots)
    calculated_intensities = gsd.find(spots)
    codebook.decode_euclidean(calculated_intensities)

    # spots are detected in a different order that they're generated; sorting makes comparison easy
    sorted_intensities = intensities.sortby('features')
    sorted_calculated_intensities = calculated_intensities.sortby('features')

    # verify that the spots are all detected, and decode to the correct genes
    assert np.array_equal(
        sorted_intensities.coords[intensities.Constants.GENE],
        sorted_calculated_intensities.coords[intensities.Constants.GENE]
    )



from typing import Union

import numpy as np
import pandas as pd
import xarray as xr

from starfish.constants import Indices, AugmentedEnum


class IntensityTable(xr.DataArray):
    """3 dimensional container for spot/pixel features extracted from image data

    An IntensityTable is comprised of each feature's intensity across channels and hybridization
    rounds, forming a (feature, channel, hybridization) tensor implemented as an xarray.DataArray
    object. In addition to the basic xarray methods, IntensityTable implements:

    Constructors
    -------
    empty_intensity_table  creates an IntensityTable with all intensities equal to zero
    from_spot_data         creates an IntensityTable from a 3d array and a spot attributes dataframe
    synthetic_intensities  creates an IntensityTable with synthetic spots, given a codebook

    Methods
    -------
    save                   save the IntensityTable to netCDF
    load                   load an IntensityTable from netCDF

    Attributes
    ----------
    Constants.FEATURES     name of the first axis of the IntensityTable
    Constants.GENE :       name of the field that stores the decoded gene identity for each feature
    Constants.QUALITY      name of the field that stores the decoded gene quality for each feature
    SpotAttributes.X       name of the pixelwise spot x-coordinate
    SpotAttributes.Y       name of the pixelwise spot y-coordinate
    SpotAttributes.Z       name of the pixelwise spot z-coordinate
    SpotAttributes.RADIUS  name of the spot radius index

    """

    class Constants(AugmentedEnum):
        FEATURES = 'features'
        GENE = 'gene_name'
        QUALITY = 'quality'

    class SpotAttributes(AugmentedEnum):
        X = 'x'
        Y = 'y'
        Z = 'z'
        RADIUS = 'r'

    @classmethod
    def empty_intensity_table(
            cls, spot_attributes: pd.MultiIndex, n_ch: int, n_hyb: int
    ) -> "IntensityTable":
        """Create an empty intensity table with pre-set axis whose values are zero

        Parameters
        ----------
        spot_attributes : pd.MultiIndex
            MultiIndex containing spot metadata. Must contain the values specifid in Constants.X,
            Y, Z, and RADIUS.
        n_ch : int
            number of channels measured in the imaging experiment
        n_hyb : int
            number of hybridization rounds measured in the imaging experiment

        Returns
        -------
        IntensityTable :
            empty IntensityTable

        """
        cls._verify_spot_attributes(spot_attributes)
        channel_index = pd.Index(np.arange(n_ch))
        hyb_index = pd.Index(np.arange(n_hyb))
        data = np.zeros((spot_attributes.shape[0], n_ch, n_hyb))
        dims = (IntensityTable.Constants.FEATURES.value, Indices.CH.value, Indices.HYB.value)
        return cls(data=data, coords=(spot_attributes, channel_index, hyb_index), dims=dims)

    @staticmethod
    def _verify_spot_attributes(spot_attributes: pd.MultiIndex) -> None:
        """Run some checks on spot attributes"""
        if not isinstance(spot_attributes, pd.MultiIndex):
            raise ValueError(
                f'spot attributes must be a pandas MultiIndex, not {type(spot_attributes)}.')

        required_attributes = set(a.value for a in IntensityTable.SpotAttributes)
        missing_attributes = required_attributes.difference(spot_attributes.names)
        if missing_attributes:
            raise ValueError(
                f'Missing spot_attribute levels in provided MultiIndex: {missing_attributes}. '
                f'The following levels are required: {required_attributes}.')

    @classmethod
    def from_spot_data(
            cls, intensities: Union[xr.DataArray, np.ndarray], spot_attributes: pd.MultiIndex,
            *args, **kwargs) -> "IntensityTable":
        """Table to store image feature intensities and associated metadata

        Parameters
        ----------
        intensities : np.ndarray[Any]
            intensity data
        spot_attributes : pd.MultiIndex
            Name(s) of the data dimension(s). Must be either a string (only
            for 1D data) or a sequence of strings with length equal to the
            number of dimensions. If this argument is omitted, dimension names
            are taken from ``coords`` (if possible) and otherwise default to
            ``['dim_0', ... 'dim_n']``.
        args :
            additional arguments to pass to the xarray constructor
        kwargs :
            additional keyword arguments to pass to the xarray constructor

        Returns
        -------
        IntensityTable :
            IntensityTable containing data from passed ndarray, annotated by spot_attributes

        """

        if len(intensities.shape) != 3:
            raise ValueError(
                f'intensities must be a (features * ch * hyb) 3-d tensor. Provided intensities '
                f'shape ({intensities.shape}) is invalid.')

        cls._verify_spot_attributes(spot_attributes)

        coords = (
            (IntensityTable.Constants.FEATURES.value, spot_attributes),
            (Indices.CH.value, np.arange(intensities.shape[1])),
            (Indices.HYB.value, np.arange(intensities.shape[2]))
        )

        dims = (IntensityTable.Constants.FEATURES.value, Indices.CH.value, Indices.HYB.value)

        return cls(intensities, coords, dims, *args, **kwargs)

    def save(self, filename: str) -> None:
        """Save an IntensityTable as a Netcdf File

        Parameters
        ----------
        filename : str
            Name of Netcdf file

        """
        # TODO when https://github.com/pydata/xarray/issues/1077 (support for multiindex
        # serliazation) is merged, remove this reset_index() call and simplify load, below
        self.reset_index('features').to_netcdf(filename)

    @classmethod
    def load(cls, filename: str) -> "IntensityTable":
        """load an IntensityTable from Netcdf

        Parameters
        ----------
        filename : str
            File to load

        Returns
        -------
        IntensityTable

        """
        loaded = xr.open_dataarray(filename)
        intensity_table = cls(
            loaded.data,
            loaded.coords,
            loaded.dims
        )
        return intensity_table.set_index(features=list(intensity_table['features'].coords.keys()))

    def show(self, background_image: np.ndarray) -> None:
        """show spots on a background image"""
        raise NotImplementedError

    @classmethod
    def synthetic_intensities(
            cls, codebook, num_z: int=12, height: int=50, width: int=40, n_spots=10,
            mean_fluor_per_spot=200, mean_photons_per_fluor=50
    ) -> "IntensityTable":
        """Create an IntensityTable containing synthetic spots with random locations

        Parameters
        ----------
        codebook : Codebook
            starfish codebook object
        num_z :
            number of z-planes to use when localizing spots
        height :
            y dimension of each synthetic plane
        width :
            x dimension of each synthetic plane
        n_spots :
            number of spots to generate
        mean_fluor_per_spot :
            mean number of fluorophores per spot
        mean_photons_per_fluor :
            mean number of photons per fluorophore.

        Returns
        -------
        IntensityTable

        """

        # TODO nsofroniew: right now there is no jitter on x-y positions of the spots
        z = np.random.randint(0, num_z, size=n_spots)
        y = np.random.randint(0, height, size=n_spots)
        x = np.random.randint(0, width, size=n_spots)
        r = np.empty(n_spots)
        r.fill(np.nan)  # radius is a function of the point-spread gaussian size

        names = [cls.SpotAttributes.Z.value, cls.SpotAttributes.Y.value,
                 cls.SpotAttributes.X.value, cls.SpotAttributes.RADIUS.value]
        spot_attributes = pd.MultiIndex.from_arrays([z, y, x, r], names=names)

        # empty data tensor
        data = np.zeros(shape=(n_spots, *codebook.shape[1:]))

        genes = np.random.choice(
            codebook.coords[cls.Constants.GENE.value], size=n_spots, replace=True)
        expected_bright_locations = np.where(codebook.loc[genes])

        # create a binary matrix where "on" spots are 1
        data[expected_bright_locations] = 1

        # add physical properties of fluorescence
        data *= np.random.poisson(mean_photons_per_fluor, size=data.shape)
        data *= np.random.poisson(mean_fluor_per_spot, size=data.shape)

        intensities = cls.from_spot_data(data, spot_attributes)
        intensities[cls.Constants.GENE.value] = ('features', genes)

        return intensities



import warnings

import numpy
import pytest

from starfish.errors import DataFormatWarning
from starfish.test.dataset_fixtures import synthetic_stack


NUM_HYB = 2
NUM_CH = 2
NUM_Z = 2
HEIGHT = 10
WIDTH = 10


def create_tile_data_provider(dtype: numpy.number, corner_dtype: numpy.number):
    """
    Makes a stack that's all of the same type, except the hyb=0,ch=0,z=0 corner, which is a different type.  All the
    tiles are initialized with ones.

    Parameters
    ----------
    dtype : numpy.number
        The data type of all the tiles except the hyd=0,ch=0,z=0 corner.
    corner_dtype
        The data type of the tile in the hyd=0,ch=0,z=0 corner.

    Returns
    -------
    ImageStack :
        The image stack with the tiles initialized as described.
    """
    def tile_data_provider(hyb: int, ch: int, z: int, height: int, width: int) -> numpy.ndarray:
        if hyb == 0 and ch == 0 and z == 0:
            return numpy.ones((height, width), dtype=corner_dtype)
        else:
            return numpy.ones((height, width), dtype=dtype)
    return tile_data_provider


def test_multiple_tiles_of_different_kind():
    with pytest.raises(TypeError):
        synthetic_stack(
            NUM_HYB, NUM_CH, NUM_Z,
            HEIGHT, WIDTH,
            tile_data_provider=create_tile_data_provider(numpy.uint32, numpy.float32),
        )


def test_multiple_tiles_of_same_dtype():
    stack = synthetic_stack(
        NUM_HYB, NUM_CH, NUM_Z,
        HEIGHT, WIDTH,
        tile_data_provider=create_tile_data_provider(numpy.uint32, numpy.uint32),
    )
    expected = numpy.ones(
        (NUM_HYB,
         NUM_CH,
         NUM_Z,
         HEIGHT,
         WIDTH), dtype=numpy.uint32)
    assert numpy.array_equal(stack.numpy_array, expected)


def test_int_type_promotion():
    with warnings.catch_warnings(record=True) as w:
        stack = synthetic_stack(
            NUM_HYB, NUM_CH, NUM_Z,
            HEIGHT, WIDTH,
            tile_data_provider=create_tile_data_provider(numpy.int32, numpy.int8),
        )
        assert len(w) == 1
        assert issubclass(w[0].category, DataFormatWarning)
    expected = numpy.ones(
        (NUM_HYB,
         NUM_CH,
         NUM_Z,
         HEIGHT,
         WIDTH), dtype=numpy.int32)
    corner = numpy.empty(
        (HEIGHT,
         WIDTH), dtype=numpy.int32)
    corner.fill(16777216)
    expected[0, 0, 0] = corner
    assert numpy.array_equal(stack.numpy_array, expected)


def test_uint_type_promotion():
    with warnings.catch_warnings(record=True) as w:
        stack = synthetic_stack(
            NUM_HYB, NUM_CH, NUM_Z,
            HEIGHT, WIDTH,
            tile_data_provider=create_tile_data_provider(numpy.uint32, numpy.uint8),
        )
        assert len(w) == 1
        assert issubclass(w[0].category, DataFormatWarning)
    expected = numpy.ones(
        (NUM_HYB,
         NUM_CH,
         NUM_Z,
         HEIGHT,
         WIDTH), dtype=numpy.uint32)
    corner = numpy.empty(
        (HEIGHT,
         WIDTH), dtype=numpy.uint32)
    corner.fill(16777216)
    expected[0, 0, 0] = corner
    assert numpy.array_equal(stack.numpy_array, expected)


def test_float_type_promotion():
    with warnings.catch_warnings(record=True) as w:
        stack = synthetic_stack(
            NUM_HYB, NUM_CH, NUM_Z,
            HEIGHT, WIDTH,
            tile_data_provider=create_tile_data_provider(numpy.float64, numpy.float32),
        )
        assert len(w) == 1
        assert issubclass(w[0].category, DataFormatWarning)
    expected = numpy.ones(
        (NUM_HYB,
         NUM_CH,
         NUM_Z,
         HEIGHT,
         WIDTH), dtype=numpy.float64)
    assert numpy.array_equal(stack.numpy_array, expected)



from functools import partial

import numpy
from skimage.filters import gaussian

from starfish.image import ImageStack
from ._base import FilterAlgorithmBase


# TODO ambrosejcarr: need a better solution for 2d/3d support for image analysis algorithms

class GaussianLowPass(FilterAlgorithmBase):

    def __init__(self, sigma, is_volume: bool=False, verbose=False, **kwargs) -> None:
        """Multi-dimensional low-pass gaussian filter.

        Parameters
        ----------
        sigma : Union[float, Tuple[float]]
            Standard deviation for Gaussian kernel.
        is_volume : bool
            If True, 3d (z, y, x) volumes will be filtered, otherwise, filter 2d tiles independently.
        verbose : bool
            If True, report on the percentage completed (default = False) during processing

        """
        self.sigma = sigma
        self.is_volume = is_volume
        self.verbose = verbose

    @classmethod
    def add_arguments(cls, group_parser):
        group_parser.add_argument(
            "--sigma", default=1, type=int, help="standard deviation of gaussian kernel")

    @staticmethod
    def low_pass(image, sigma) -> numpy.ndarray:
        """Apply a Gaussian blur operation over a multi-dimensional image.

        Parameters
        ----------
        image : np.ndarray
            Image data
        sigma : Union[float, int, Tuple]
            Standard deviation of the Gaussian kernel that will be applied. If a float, an isotropic kernel will be
            assumed, otherwise the dimensions of the kernel give (z, x, y)

        Returns
        -------
        np.ndarray :
            Blurred data in same shape as input image.

        """
        blurred = gaussian(
            image, sigma=sigma, output=None, cval=0, multichannel=True, preserve_range=True, truncate=4.0)

        blurred = blurred.astype(numpy.uint16)

        return blurred

    def filter(self, stack: ImageStack) -> None:
        """Perform in-place filtering of an image stack and all contained aux images.

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.

        """
        low_pass = partial(self.low_pass, sigma=self.sigma)
        stack.apply(low_pass, is_volume=self.is_volume, verbose=self.verbose)



from typing import Tuple

import numpy as np
import pandas as pd
from skimage.feature import blob_log

from starfish.constants import Indices
from starfish.image import ImageStack
from starfish.munge import melt
from starfish.pipeline.features.encoded_spots import EncodedSpots
from starfish.pipeline.features.spot_attributes import SpotAttributes
from starfish.util.argparse import FsExistsType
from ._base import SpotFinderAlgorithmBase


class GaussianSpotDetector(SpotFinderAlgorithmBase):

    def __init__(
            self, min_sigma, max_sigma, num_sigma, threshold,
            blobs_stack, overlap=0.5, measurement_type='max', **kwargs):
        """Multi-dimensional gaussian spot detector

        Parameters
        ----------
        min_sigma : float
            The minimum standard deviation for Gaussian Kernel. Keep this low to
            detect smaller blobs.
        max_sigma : float
            The maximum standard deviation for Gaussian Kernel. Keep this high to
            detect larger blobs.
        num_sigma : int
            The number of intermediate values of standard deviations to consider
            between `min_sigma` and `max_sigma`.
        threshold : float
            The absolute lower bound for scale space maxima. Local maxima smaller
            than thresh are ignored. Reduce this to detect blobs with less
            intensities.
        overlap : float [0, 1]
            If two spots have more than this fraction of overlap, the spots are combined (default = 0.5)
        blobs_stack : Union[ImageStack, str]
            ImageStack or the path or URL that references the ImageStack that contains the blobs.
        measurement_type : str ['max', 'mean']
            name of the function used to calculate the intensity for each identified spot area

        Notes
        -----
        This spot detector is very sensitive to the threshold that is selected, and the threshold is defined as an
        absolute value -- therefore it must be adjusted depending on the datatype of the passed image.


        """
        self.min_sigma = min_sigma
        self.max_sigma = max_sigma
        self.num_sigma = num_sigma
        self.threshold = threshold
        self.overlap = overlap
        if isinstance(blobs_stack, ImageStack):
            self.blobs_stack = blobs_stack
        else:
            self.blobs_stack = ImageStack.from_path_or_url(blobs_stack)

        try:
            self.measurement_function = getattr(np, measurement_type)
        except AttributeError:
            raise ValueError(
                f'measurement_type must be a numpy reduce function such as "max" or "mean". {measurement_type} '
                f'not found.')

    @staticmethod
    def measure_blob_intensity(image, spots, measurement_function) -> pd.Series:
        def fn(row):
            x_min = int(round(row.x_min))
            x_max = int(round(row.x_max))
            y_min = int(round(row.y_min))
            y_max = int(round(row.y_max))
            return measurement_function(image[x_min:x_max, y_min:y_max])
        return spots.apply(
            fn,
            axis=1
        )

    def encode(self, stack, spot_attributes):
        # create stack squeeze map
        squeezed = stack.squeeze()
        mapping: pd.DataFrame = stack.tile_metadata
        inds = range(mapping.shape[0])
        intensities = [
            self.measure_blob_intensity(image, spot_attributes, self.measurement_function)
            for image in squeezed
        ]
        d = dict(zip(inds, intensities))
        d['spot_id'] = range(spot_attributes.shape[0])

        res = pd.DataFrame(d)

        res: pd.DataFrame = melt(
            df=res,
            new_index_name='barcode_index',
            new_value_name='intensity',
            melt_columns=inds
        )
        res = pd.merge(res, mapping, on='barcode_index', how='left')
        return EncodedSpots(res)

    def fit(self, blobs_image):
        fitted_blobs = pd.DataFrame(
            data=blob_log(blobs_image, self.min_sigma, self.max_sigma, self.num_sigma, self.threshold, self.overlap),
            columns=['x', 'y', 'r'],
        )

        # TODO ambrosejcarr: why is this necessary? (check docs)
        fitted_blobs['r'] *= np.sqrt(2)
        fitted_blobs[['x', 'y']] = fitted_blobs[['x', 'y']].astype(int)

        fitted_blobs['x_min'] = np.clip(np.floor(fitted_blobs.x - fitted_blobs.r), 0, None)
        fitted_blobs['x_max'] = np.clip(np.ceil(fitted_blobs.x + fitted_blobs.r), None, blobs_image.shape[0])
        fitted_blobs['y_min'] = np.clip(np.floor(fitted_blobs.y - fitted_blobs.r), 0, None)
        fitted_blobs['y_max'] = np.clip(np.ceil(fitted_blobs.y + fitted_blobs.r), None, blobs_image.shape[1])

        # TODO ambrosejcarr this should be barcode intensity or position intensity
        fitted_blobs['intensity'] = self.measure_blob_intensity(blobs_image, fitted_blobs, self.measurement_function)
        fitted_blobs['spot_id'] = np.arange(fitted_blobs.shape[0])

        return SpotAttributes(fitted_blobs)

    def find(self, hybridization_image) -> Tuple[SpotAttributes, EncodedSpots]:
        blobs = self.blobs_stack.max_proj(Indices.HYB, Indices.CH, Indices.Z)
        spot_attributes = self.fit(blobs)
        encoded_spots = self.encode(hybridization_image, spot_attributes.data)
        return spot_attributes, encoded_spots

    @classmethod
    def add_arguments(cls, group_parser):
        group_parser.add_argument("--blobs-stack", type=FsExistsType(), required=True)
        group_parser.add_argument(
            "--min-sigma", default=4, type=int, help="Minimum spot size (in standard deviation)")
        group_parser.add_argument(
            "--max-sigma", default=6, type=int, help="Maximum spot size (in standard deviation)")
        group_parser.add_argument("--num-sigma", default=20, type=int, help="Number of scales to try")
        group_parser.add_argument("--threshold", default=.01, type=float, help="Dots threshold")
        group_parser.add_argument(
            "--overlap", default=0.5, type=float, help="dots with overlap of greater than this fraction are combined")
        group_parser.add_argument(
            "--show", default=False, action='store_true', help="display results visually")



import argparse
import os


class FsExistsType:
    def __call__(self, prospective_dir):
        if not os.path.exists(prospective_dir):
            raise argparse.ArgumentTypeError("{0} does not exist".format(prospective_dir))
        return prospective_dir



from starfish.pipeline.algorithmbase import AlgorithmBase


class GeneAssignmentAlgorithm(AlgorithmBase):
    def assign_genes(self, spots, regions):
        """Performs gene assignment given the spots and the regions."""
        raise NotImplementedError()



from typing import Tuple

from starfish.image import ImageStack
from starfish.pipeline.algorithmbase import AlgorithmBase
from starfish.pipeline.features.spot_attributes import SpotAttributes
from starfish.pipeline.features.encoded_spots import EncodedSpots


class SpotFinderAlgorithmBase(AlgorithmBase):
    def find(self, hybridization_image: ImageStack) -> Tuple[SpotAttributes, EncodedSpots]:
        """Find spots."""
        raise NotImplementedError()



import argparse
import os

from starfish.image import ImageStack
from starfish.pipeline.pipelinecomponent import PipelineComponent
from starfish.util.argparse import FsExistsType
from . import _base
from . import gaussian
from . import local_max_peak_finder


class SpotFinder(PipelineComponent):

    spot_finder_group: argparse.ArgumentParser

    @classmethod
    def implementing_algorithms(cls):
        return _base.SpotFinderAlgorithmBase.__subclasses__()

    @classmethod
    def add_to_parser(cls, subparsers):
        """Adds the spot finder component to the CLI argument parser."""
        spot_finder_group = subparsers.add_parser("detect_spots")
        spot_finder_group.add_argument("-i", "--input", type=FsExistsType(), required=True)
        spot_finder_group.add_argument("-o", "--output", required=True)
        spot_finder_group.set_defaults(starfish_command=SpotFinder._cli)
        spot_finder_subparsers = spot_finder_group.add_subparsers(dest="spot_finder_algorithm_class")

        for algorithm_cls in cls.algorithm_to_class_map().values():
            group_parser = spot_finder_subparsers.add_parser(algorithm_cls.get_algorithm_name())
            group_parser.set_defaults(spot_finder_algorithm_class=algorithm_cls)
            algorithm_cls.add_arguments(group_parser)

        cls.spot_finder_group = spot_finder_group

    @classmethod
    def _cli(cls, args, print_help=False):
        """Runs the spot finder component based on parsed arguments."""

        if args.spot_finder_algorithm_class is None or print_help:
            cls.spot_finder_group.print_help()
            cls.spot_finder_group.exit(status=2)

        print('Detecting Spots ...')
        hybridization_stack = ImageStack.from_path_or_url(args.input)
        instance = args.spot_finder_algorithm_class(**vars(args))
        spot_attributes, encoded_spots = instance.find(hybridization_stack)

        if args.show:
            encoded_spots.show(figsize=(10, 10))

        path = os.path.join(args.output, 'spots.geojson')
        print(f"Writing | spots geojson to: {path}")
        spot_attributes.save_geojson(path)

        path = os.path.join(args.output, 'spots.json')
        print(f"Writing | spot_id | x | y | z | to: {path}")
        spot_attributes.save(path)

        path = os.path.join(args.output, 'encoder_table.json')
        print(f"Writing | spot_id | hyb | ch | val | to: {path}")
        encoded_spots.save(path)



from ._base import DecoderAlgorithmBase
from starfish.constants import Indices
from starfish.pipeline.features.decoded_spots import DecodedSpots


class IssDecoder(DecoderAlgorithmBase):
    def __init__(self, **kwargs):
        pass

    @classmethod
    def add_arguments(cls, group_parser):
        pass

    def decode(self, encoded, codebook, letters=('T', 'G', 'C', 'A')):
        import numpy as np
        import pandas as pd

        num_ch = encoded[Indices.CH.value].max() + 1
        num_hyb = encoded[Indices.HYB.value].max() + 1
        num_spots = encoded['spot_id'].max() + 1

        seq_res = np.zeros((num_spots, num_hyb))
        seq_stren = np.zeros((num_spots, num_hyb))
        seq_qual = np.zeros((num_spots, num_hyb))

        for spot_id in range(num_spots):

            sid_df = encoded[encoded.spot_id == spot_id]

            mat = np.zeros((num_hyb, num_ch))
            inds = zip(sid_df[Indices.HYB.value].values, sid_df[Indices.CH.value].values, sid_df['intensity'])

            for tup in inds:
                mat[tup[0], tup[1]] = tup[2]

            max_stren = np.max(mat, axis=1)
            max_ind = np.argmax(mat, axis=1)
            qual = max_stren / np.sum(mat, axis=1)

            seq_res[spot_id, :] = max_ind
            seq_stren[spot_id, :] = max_stren
            seq_qual[spot_id, :] = qual

        max_qual = np.max(seq_qual, axis=1)

        codes = []
        for k in range(seq_res.shape[0]):
            letter_inds = seq_res[k, :]
            letter_inds = letter_inds.astype(np.int)
            res = ''.join([letters[j] for j in letter_inds])
            codes.append(res)

        dec = pd.DataFrame({'spot_id': range(num_spots),
                            'barcode': codes,
                            'quality': max_qual})

        dec = pd.merge(dec, codebook, on='barcode', how='left')

        return DecodedSpots(dec)



from functools import partial

import numpy

from starfish.image import ImageStack
from ._base import FilterAlgorithmBase


class Clip(FilterAlgorithmBase):

    def __init__(self, p_min, p_max, is_volume: bool=False, verbose: bool=False, **kwargs) -> None:
        """Image clipping filter

        Parameters
        ----------
        p_min : int
            values below this percentile are set to p_min
        p_max : int
            values above this percentile are set to p_max
        is_volume : bool
            If True, 3d (z, y, x) volumes will be passed to self.low_pass. By default, 2d tiles will be passed
        verbose : bool
            If True, report on the percentage completed (default = False) during processing
        kwargs
        """
        self.p_min: int = p_min
        self.p_max: int = p_max
        self.is_volume = is_volume
        self.verbose = verbose

    @classmethod
    def add_arguments(cls, group_parser):
        group_parser.add_argument("--p-min", default=0, type=int, help="clip intensities below this percentile")
        group_parser.add_argument("--p-max", default=100, type=int, help="clip intensities above this percentile")

    @staticmethod
    def clip(image, p_min: int, p_max: int) -> numpy.ndarray:
        """Clip values of img below and above percentiles p_min and p_max

        Parameters
        ----------
        p_min, int (default 0)
          values below this percentile are set to the value of this percentile
        p_max, int (default 100)
          values above this percentile are set to the value of this percentile

        Notes
        -----
        - Wrapper for np.clip
        - No shifting or transformation to adjust dynamic range is done after
          clipping

        Returns
        -------
        np.ndarra
          Numpy array of same shape as img

        """
        v_min, v_max = numpy.percentile(image, [p_min, p_max])

        # asking for a float percentile clipping value from an integer image will
        # convert to float, so store the dtype so it can be restored
        dtype = image.dtype
        image = image.clip(min=v_min, max=v_max)
        return image.astype(dtype)

    def filter(self, stack: ImageStack) -> None:
        """Perform in-place filtering of an image stack and all contained aux images.

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.

        """
        clip = partial(self.clip, p_min=self.p_min, p_max=self.p_max)
        stack.apply(clip, is_volume=self.is_volume, verbose=self.verbose)



import argparse

from starfish.image import ImageStack
from starfish.pipeline.pipelinecomponent import PipelineComponent
from starfish.util.argparse import FsExistsType
from . import _base
from . import bandpass
from . import clip
from . import gaussian_high_pass
from . import gaussian_low_pass
from . import richardson_lucy_deconvolution
from . import white_tophat


class Filter(PipelineComponent):

    filter_group: argparse.ArgumentParser

    @classmethod
    def implementing_algorithms(cls):
        return _base.FilterAlgorithmBase.__subclasses__()

    @classmethod
    def add_to_parser(cls, subparsers):
        """Adds the filter component to the CLI argument parser."""
        filter_group = subparsers.add_parser("filter")
        filter_group.add_argument("-i", "--input", type=FsExistsType(), required=True)
        filter_group.add_argument("-o", "--output", required=True)
        filter_group.set_defaults(starfish_command=Filter._cli)
        filter_subparsers = filter_group.add_subparsers(dest="filter_algorithm_class")

        for algorithm_cls in cls.algorithm_to_class_map().values():
            group_parser = filter_subparsers.add_parser(algorithm_cls.get_algorithm_name())
            group_parser.set_defaults(filter_algorithm_class=algorithm_cls)
            algorithm_cls.add_arguments(group_parser)

        cls.filter_group = filter_group

    @classmethod
    def _cli(cls, args, print_help=False):
        """Runs the filter component based on parsed arguments."""

        if args.filter_algorithm_class is None or print_help:
            cls.filter_group.print_help()
            cls.filter_group.exit(status=2)

        print('Filtering images ...')
        stack = ImageStack.from_path_or_url(args.input)
        instance = args.filter_algorithm_class(**vars(args))
        instance.filter(stack)

        stack.write(args.output)



import json
import os
from typing import Optional, Mapping

import numpy as np
from slicedimage import TileSet, Tile
from slicedimage.io import resolve_path_or_url

from starfish.constants import Coordinates, Indices
from .image import ImageStack


class Stack:

    def __init__(self):
        # data organization
        self.org = None
        self.image = None

        # auxiliary images
        self.auxiliary_images = dict()

        # readers and writers
        self.write_fn = np.save  # asserted for now

        # backend & baseurl
        self.backend = None
        self.baseurl = None

    @classmethod
    def from_data(cls, image_stack: ImageStack, aux_dict: Optional[Mapping[str, ImageStack]]=None) -> "Stack":
        """create a Stack from an already-loaded ImageStack

        Parameters
        ----------
        image_stack : ImageStack
            in-memory ImageStack
        aux_dict : Optional[Mapping[str, ImageStack]]
            a dictionary of ImageStacks, default None

        Returns
        -------
        Stack :
            a Stack object

        """
        stack = cls()
        stack.image = image_stack
        stack.auxiliary_images = aux_dict if aux_dict is not None else dict()
        return stack

    def read(self, in_json_path_or_url):
        self.backend, name, self.baseurl = resolve_path_or_url(in_json_path_or_url)
        with self.backend.read_file_handle(name) as fh:
            self.org = json.load(fh)

        self.image = ImageStack.from_url(self.org['hybridization_images'], self.baseurl)
        self._read_aux()

    def _read_aux(self):
        for aux_key, aux_data in self.org['auxiliary_images'].items():
            self.auxiliary_images[aux_key] = ImageStack.from_url(aux_data, self.baseurl)

    @classmethod
    def from_experiment_json(cls, json_url: str) -> "Stack":
        """Construct a `Stack` from an experiment.json file format specifier

        Parameters
        ----------
        json_url : str
            file path or web link to an experiment.json file

        Returns
        -------
        Stack :
            Stack object serving the requested image data

        """
        stack: Stack = cls()
        stack.read(json_url)
        return stack

    # TODO should this thing write npy?
    def write(self, dir_name):
        self._write_stack(dir_name)
        self._write_aux(dir_name)
        self._write_metadata(dir_name)

    def _write_metadata(self, dir_name):
        with open(os.path.join(dir_name, 'experiment.json'), 'w') as outfile:
            json.dump(self.org, outfile, indent=4)

    def _write_stack(self, dir_name):
        stack_path = os.path.join(dir_name, "hybridization.json")
        self.image.write(stack_path)
        self.org['hybridization_images'] = "hybridization.json"

    def _write_aux(self, dir_name):
        for aux_key, aux_data in self.org['auxiliary_images'].items():
            self.auxiliary_images[aux_key].write(os.path.join(dir_name, aux_data))

    def set_stack(self, new_stack):
        if self.image.raw_shape != new_stack.shape:
            msg = "Shape mismatch. Current data shape: {}, new data shape: {}".format(
                self.image.shape, new_stack.shape)
            raise AttributeError(msg)
        self.image.numpy_array = new_stack

    def set_aux(self, key, img):
        if key in self.auxiliary_images:
            old_img = self.auxiliary_images[key]
            if old_img.shape != img.shape:
                msg = "Shape mismatch. Current data shape: {}, new data shape: {}".format(
                    old_img.shape, img.shape)
                raise AttributeError(msg)
            self.auxiliary_images[key].numpy_array = img
        else:
            # TODO: (ttung) major hack alert.  we don't have a convenient mechanism to build an ImageStack from a single
            # numpy array, which we probably should.
            tileset = TileSet(
                {
                    Indices.HYB,
                    Indices.CH,
                    Indices.Z,
                    Coordinates.X,
                    Coordinates.Y,
                },
                {
                    Indices.HYB: 1,
                    Indices.CH: 1,
                    Indices.Z: 1,
                }
            )
            tile = Tile(
                {
                    Coordinates.X: (0.000, 0.001),
                    Coordinates.Y: (0.000, 0.001),
                },
                {
                    Indices.HYB: 0,
                    Indices.CH: 0,
                    Indices.Z: 0,
                },
                img.shape,
            )
            tile.numpy_array = img
            tileset.add_tile(tile)

            self.auxiliary_images[key] = ImageStack(tileset)
            self.org['auxiliary_images'][key] = f"{key}.json"



import functools
import json
from typing import Any, Dict, List, Optional, Sequence, Set
import uuid

import numpy as np
import pandas as pd
import xarray as xr

from starfish.constants import Indices, AugmentedEnum
from starfish.pipeline.features.intensity_table import IntensityTable


class Codebook(xr.DataArray):
    """Codebook for a spatial transcriptomics experiment

    The codebook is a three dimensional tensor whose values are the expected intensity of a spot
    for each code in each hybridization round and each color channel. This class supports the
    construction of synthetic codebooks for testing, and exposes decode methods to assign gene
    identifiers to spots. This codebook provides an in-memory representation of the codebook
    defined in the spaceTx format.

    The codebook is a subclass of xarray, and exposes the complete public API of that package in
    addition to the methods and constructors listed below.

    Constructors
    ------------
    from_code_array(code_array, n_hyb, n_ch)
        construct a codebook from a spaceTx-spec array of codewords
    from_json(json_codebook, n_hyb, n_ch)
        load a codebook from a spaceTx spec-compliant json file
    synthetic_one_hot_codebook
        construct a codebook of random codes where only one channel is on per hybridization round

    Methods
    -------
    decode_euclidean(intensities)
        find the closest code for each spot in intensities by euclidean distance
    decode_per_channel_maximum(intensities)
        find codes that match the per-channel max intensity for each spot in intensities
    code_length()
        return the total length of the codes in the codebook

    Attributes
    ----------
    Constants.CODEWORD
        name of codeword field in spaceTx spec
    Constants.GENE
        name of gene specifier in spaceTx spec
    Constants.VALUE
        name of value specifier in SpaceTx spec

    See Also
    --------
    TODO <link to spaceTx format>

    """

    class Constants(AugmentedEnum):
        CODEWORD = 'codeword'
        GENE = 'gene_name'
        VALUE = 'v'

    @property
    def code_length(self) -> int:
        """return the length of codes in this codebook"""
        return int(np.dot(*self.shape[1:]))

    @classmethod
    def _empty_codebook(cls, code_names: Sequence[str], n_ch: int, n_hyb: int):
        """create an empty codebook of shape (code_names, n_ch, n_hyb)

        Parameters
        ----------
        code_names : Sequence[str]
            the genes to be coded
        n_ch : int
            number of channels used to build the codes
        n_hyb : int
            number of hybridization rounds used to build the codes

        Returns
        -------
        Codebook :
            codebook whose values are all zero

        """
        codes_index = pd.Index(code_names, name=Codebook.Constants.GENE.value)
        return cls(
            data=np.zeros((codes_index.shape[0], n_ch, n_hyb), dtype=np.uint8),
            coords=(
                codes_index,
                pd.Index(np.arange(n_ch), name=Indices.CH.value),
                pd.Index(np.arange(n_hyb), name=Indices.HYB.value),
            )
        )

    @classmethod
    def from_code_array(
            cls, code_array: List[Dict[str, Any]],
            n_hyb: Optional[int]=None, n_ch: Optional[int]=None) -> "Codebook":
        """construct a codebook from a spaceTx-spec array of codewords

        Parameters
        ----------
        code_array : List[Dict[str, Any]]
            Array of dictionaries, each containing a codeword and gene_name
        n_hyb : Optional[int]
            The number of hybridization rounds used in the codes. Will be inferred if not provided
        n_ch : Optional[int]
            The number of channels used in the codes. Will be inferred if not provided

        Returns
        -------
        Codebook :
            Codebook with shape (genes, channels, hybridization_rounds)

        """

        # guess the max hyb and channel if not provided, otherwise check provided values are valid
        max_hyb, max_ch = 0, 0

        for code in code_array:
            for entry in code[Codebook.Constants.CODEWORD.value]:
                max_hyb = max(max_hyb, entry[Indices.HYB])
                max_ch = max(max_ch, entry[Indices.CH])

        # set n_ch and n_hyb if either were not provided
        n_hyb = n_hyb if n_hyb is not None else max_hyb + 1
        n_ch = n_ch if n_ch is not None else max_ch + 1

        # raise errors if provided n_hyb or n_ch are out of range
        if max_hyb + 1 > n_hyb:
            raise ValueError(
                f'code detected that requires a hybridization value ({max_hyb}) that is greater '
                f'than provided n_hyb: {n_hyb}')
        if max_ch + 1 > n_ch:
            raise ValueError(
                f'code detected that requires a hybridization value ({max_hyb}) that is greater '
                f'than provided n_hyb: {n_hyb}')

        # verify codebook structure and fields
        for code in code_array:

            if not isinstance(code, dict):
                raise ValueError(f'codebook must be an array of dictionary codes. Found: {code}.')

            # verify all necessary fields are present
            required_fields = {Codebook.Constants.CODEWORD.value, Codebook.Constants.GENE.value}
            missing_fields = required_fields.difference(code)
            if missing_fields:
                raise ValueError(
                    f'Each entry of codebook must contain {required_fields}. Missing fields: '
                    f'{missing_fields}')

        gene_names = [w[Codebook.Constants.GENE.value] for w in code_array]
        code_data = cls._empty_codebook(gene_names, n_ch, n_hyb)

        # fill the codebook
        for code_dict in code_array:
            codeword = code_dict[Codebook.Constants.CODEWORD.value]
            gene = code_dict[Codebook.Constants.GENE.value]
            for entry in codeword:
                code_data.loc[gene, entry[Indices.CH.value], entry[Indices.HYB.value]] = entry[
                    Codebook.Constants.VALUE.value]

        return code_data

    @classmethod
    def from_json(
            cls, json_codebook: str, n_hyb: Optional[int]=None, n_ch: Optional[int]=None
    ) -> "Codebook":
        """Load a codebook from a spaceTx spec-compliant json file

        Parameters
        ----------
        json_codebook : str
            path to json file containing a spaceTx codebook
        n_hyb : Optional[int]
            The number of hybridization rounds used in the codes. Will be inferred if not provided
        n_ch : Optional[int]
            The number of channels used in the codes. Will be inferred if not provided

        Returns
        -------
        Codebook :
            Codebook with shape (genes, channels, hybridization_rounds)

        """
        with open(json_codebook, 'r') as f:
            code_array = json.load(f)
        return cls.from_code_array(code_array, n_hyb, n_ch)

    def save(self, filename: str) -> None:
        """save a codebook to json

        Notes
        -----
        This enforces the following typing of codebooks:
        ch, hyb: int
        value: float
        gene: str

        Parameters
        ----------
        filename : str
            filename

        """
        code_array = []
        for gene in self[self.Constants.GENE.value]:
            codeword = []
            for ch in self[Indices.CH.value]:
                for hyb in self[Indices.HYB.value]:
                    if self.loc[gene, ch, hyb]:
                        codeword.append(
                            {
                                Indices.CH.value: int(ch),
                                Indices.HYB.value: int(hyb),
                                self.Constants.VALUE.value: float(self.loc[gene, ch, hyb])
                            })
            code_array.append({
                self.Constants.CODEWORD.value: codeword,
                self.Constants.GENE.value: str(gene.values)
            })

        with open(filename, 'w') as f:
            json.dump(code_array, f)

    def decode_euclidean(self, intensities: IntensityTable) -> IntensityTable:
        """Assign the closest gene by euclidean distance to each feature in an intensity table

        Parameters
        ----------
        intensities : IntensityTable
            features to be decoded

        Returns
        -------
        IntensityTable :
            intensity table containing additional data variables for gene assignments and feature
            qualities

        """

        def _min_euclidean_distance(observation: xr.DataArray, codes: Codebook) -> np.ndarray:
            """find the code with the closest euclidean distance to observation

            Parameters
            ----------
            observation : xr.DataArray
                2-dimensional DataArray of shape (n_ch, n_hyb)
            codes :
                Codebook containing codes to compare to observation

            Returns
            -------
            np.ndarray :
                1-d vector containing the distance of each code to observation

            """
            squared_diff = (codes - observation) ** 2
            code_distances = np.sqrt(squared_diff.sum((Indices.CH, Indices.HYB)))
            # order of codes changes here (automated sorting on the reshaping?)
            return code_distances

        # normalize both the intensities and the codebook
        norm_intensities = intensities.groupby(IntensityTable.Constants.FEATURES.value).apply(
            lambda x: x / x.sum())
        norm_codes = self.groupby(Codebook.Constants.GENE.value).apply(lambda x: x / x.sum())

        # calculate pairwise euclidean distance between codes and features
        func = functools.partial(_min_euclidean_distance, codes=norm_codes)
        distances = norm_intensities.groupby(IntensityTable.Constants.FEATURES.value).apply(func)

        # calculate quality of each decoded spot
        qualities = 1 - distances.min(Codebook.Constants.GENE.value)
        qualities_index = pd.Index(qualities)

        # identify genes associated with closest codes
        closest_code_index = distances.argmin(Codebook.Constants.GENE.value)
        gene_ids = distances.indexes[
            Codebook.Constants.GENE.value].values[closest_code_index.values]
        gene_index = pd.Index(gene_ids)

        # set new values on the intensity table in-place
        intensities[IntensityTable.Constants.GENE.value] = (
            IntensityTable.Constants.FEATURES.value, gene_index)
        intensities[IntensityTable.Constants.QUALITY.value] = (
            IntensityTable.Constants.FEATURES.value, qualities_index)

        return intensities

    def decode_per_hyb_max(self, intensities: IntensityTable) -> IntensityTable:
        """decode each feature by selecting the per-hybridization round max-valued channel

        Notes
        -----
        If no code matches the per-channel max of a feature, it will be assigned np.nan instead
        of a gene value

        Parameters
        ----------
        intensities : IntensityTable
            features to be decoded

        Returns
        -------
        IntensityTable :
            intensity table containing additional data variables for gene assignments

        """

        def _view_row_as_element(array: np.ndarray) -> np.ndarray:
            """view an entire code as a single element

            This view allows vectors (codes) to be compared for equality without need for multiple
            comparisons by casting the data in each code to a structured dtype that registers as
            a single value

            Parameters
            ----------
            array : np.ndarray
                2-dimensional numpy array of shape (n_observations, (n_ch * n_hyb)) where
                observations may be either features or codes.

            Returns
            -------
            np.ndarray :
                1-dimensional vector of shape n_observations

            """
            nrows, ncols = array.shape
            dtype = {'names': ['f{}'.format(i) for i in range(ncols)],
                     'formats': ncols * [array.dtype]}
            return array.view(dtype)

        max_channels = intensities.argmax(Indices.CH.value)
        codes = self.argmax(Indices.CH.value)

        a = _view_row_as_element(codes.values.reshape(self.shape[0], -1))
        b = _view_row_as_element(max_channels.values.reshape(intensities.shape[0], -1))

        genes = np.empty(intensities.shape[0], dtype=object)
        genes.fill(np.nan)

        for i in np.arange(a.shape[0]):
            genes[np.where(a[i] == b)[0]] = codes['gene_name'][i]
        gene_index = pd.Index(genes.astype('U'))

        intensities[IntensityTable.Constants.GENE.value] = (
            IntensityTable.Constants.FEATURES.value, gene_index)

        return intensities

    @classmethod
    def synthetic_one_hot_codebook(
            cls, n_hyb: int, n_channel: int, n_codes: int, gene_names: Optional[Sequence]=None
    ) -> "Codebook":
        """Generate codes where one channel is "on" in each hybridization round

        Parameters
        ----------
        n_hyb : int
            number of hybridization rounds per code
        n_channel : int
            number of channels per code
        n_codes : int
            number of codes to generate
        gene_names : Optional[List[str]]
            if provided, names for genes in codebook

        Returns
        -------
        List[Dict] :
            list of codewords

        """

        # TODO ambrosejcarr: clean up this code, generate Codebooks directly using _empty_codebook
        # construct codes; this can be slow when n_codes is large and n_codes ~= n_possible_codes
        codes: Set = set()
        while len(codes) < n_codes:
            codes.add(tuple([np.random.randint(0, n_channel) for _ in np.arange(n_hyb)]))

        # construct codewords from code
        codewords = [
            [
                {
                    Indices.HYB.value: h, Indices.CH.value: c, 'v': 1
                } for h, c in enumerate(code)
            ] for code in codes
        ]

        # make a codebook from codewords
        if gene_names is None:
            # use a reverse-sorted list of integers as codewords
            gene_names = [uuid.uuid4() for _ in range(n_codes)]
        assert n_codes == len(gene_names)

        codebook = [{Codebook.Constants.CODEWORD.value: w, Codebook.Constants.GENE.value: g}
                    for w, g in zip(codewords, gene_names)]

        return cls.from_code_array(codebook, n_hyb=n_hyb, n_ch=n_channel)



import warnings
import tempfile
import os
import numpy as np
import pytest

from starfish.constants import Indices
from starfish.pipeline.features.codebook import Codebook
from starfish.pipeline.features.intensity_table import IntensityTable

# don't inspect pytest fixtures in pycharm
# noinspection PyUnresolvedReferences
from starfish.test.dataset_fixtures import (
    simple_codebook_json, simple_codebook_array, euclidean_decoded_intensities,
    per_channel_max_decoded_intensities, loaded_codebook, small_intensity_table)


def test_loading_codebook_from_json(simple_codebook_json):
    cb = Codebook.from_json(simple_codebook_json, n_ch=2, n_hyb=2)
    assert isinstance(cb, Codebook)


def test_loading_codebook_from_list(simple_codebook_array):
    cb = Codebook.from_code_array(simple_codebook_array, n_ch=2, n_hyb=2)
    assert isinstance(cb, Codebook)


def test_loading_codebook_without_specifying_ch_hyb_guesses_correct_values(simple_codebook_array):
    cb = Codebook.from_code_array(simple_codebook_array)
    assert cb.shape == (3, 2, 2)


def test_loading_codebook_with_too_few_dims_raises_value_error(simple_codebook_json):
    with pytest.raises(ValueError):
        Codebook.from_json(simple_codebook_json, n_ch=1, n_hyb=2)

    with pytest.raises(ValueError):
        Codebook.from_json(simple_codebook_json, n_ch=2, n_hyb=1)


def test_euclidean_decode_yields_correct_output(euclidean_decoded_intensities):
    expected_gene_annotation = np.array(["ACTB", "SCUBE2", "BRCA"])
    observed_gene_annotation = euclidean_decoded_intensities[
        IntensityTable.Constants.GENE.value].values
    assert np.array_equal(expected_gene_annotation, observed_gene_annotation)


def test_indexing_on_set_genes(euclidean_decoded_intensities):
    # note that this kind of indexing produces an xarray-internal FutureWarning about float
    # conversion that we can safely ignore here.
    is_actin = euclidean_decoded_intensities[IntensityTable.Constants.GENE.value] == 'ACTB'
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', FutureWarning)

        # select only the intensities that are actin, drop the rest
        result = euclidean_decoded_intensities.where(is_actin, drop=True)

    assert result.shape == (1, 2, 2)


def test_synthetic_codes_are_on_only_once_per_channel(euclidean_decoded_intensities):
    expected_gene_annotation = np.array(["ACTB", "SCUBE2", "BRCA"])
    observed_gene_annotation = euclidean_decoded_intensities[
        IntensityTable.Constants.GENE.value].values
    assert np.array_equal(expected_gene_annotation, observed_gene_annotation)


def test_per_channel_max_decode_yields_expected_results(per_channel_max_decoded_intensities):
    expected_gene_annotation = np.array(["ACTB", "SCUBE2", "BRCA"])
    observed_gene_annotation = per_channel_max_decoded_intensities[
        IntensityTable.Constants.GENE.value].values
    assert np.array_equal(expected_gene_annotation, observed_gene_annotation)


def test_synthetic_one_hot_codes_produce_one_channel_per_hyb():
    cb = Codebook.synthetic_one_hot_codebook(n_hyb=6, n_channel=4, n_codes=100)
    # sum over channels: only one should be "on"
    assert np.all(cb.sum(Indices.CH.value) == 1)


def test_codebook_save(loaded_codebook):
    directory = tempfile.mkdtemp()
    filename = os.path.join(directory, 'codebook.json')
    loaded_codebook.save(filename)
    reloaded = Codebook.from_json(filename, n_hyb=2, n_ch=2)

    assert np.array_equal(loaded_codebook, reloaded)
    assert np.array_equal(loaded_codebook[Indices.CH.value], reloaded[Indices.CH.value])
    assert np.array_equal(loaded_codebook[Indices.HYB.value], reloaded[Indices.HYB.value])
    assert np.array_equal(loaded_codebook[Codebook.Constants.GENE.value].values,
                          reloaded[Codebook.Constants.GENE.value].values)


@pytest.mark.parametrize('n_ch, n_hyb', ((2, 2), (5, 4)))
def test_loading_codebook_with_unused_channels_and_hybs(simple_codebook_json, n_ch, n_hyb):
    cb = Codebook.from_json(simple_codebook_json, n_ch=n_ch, n_hyb=n_hyb)
    assert cb.shape == (3, n_ch, n_hyb)


@pytest.mark.parametrize('n_ch, n_hyb', ((2, 2), (5, 4)))
def test_code_length(n_ch, n_hyb):
    gene_names = np.arange(10)
    cb = Codebook._empty_codebook(gene_names, n_ch, n_hyb)
    assert cb.code_length == n_ch * n_hyb



import argparse


class AlgorithmBase:
    @classmethod
    def get_algorithm_name(cls):
        """
        Returns the name of the algorithm.  This should be a valid python identifier, i.e.,
        https://docs.python.org/3/reference/lexical_analysis.html#identifiers
        """
        return cls.__name__

    @classmethod
    def add_arguments(cls, group_parser: argparse.ArgumentParser):
        """Adds the arguments for the algorithm."""
        raise NotImplementedError()



import numpy as np
import pandas as pd

from starfish.constants import Indices
from starfish.pipeline.features.intensity_table import IntensityTable
# don't inspect pytest fixtures in pycharm
# noinspection PyUnresolvedReferences
from starfish.test.dataset_fixtures import (
    loaded_codebook, simple_codebook_json, simple_codebook_array)


def test_empty_intensity_table():
    x = [1, 2]
    y = [2, 3]
    z = [1, 1]
    r = [1, 1]
    spot_attributes = pd.MultiIndex.from_arrays([x, y, z, r], names=('x', 'y', 'z', 'r'))
    empty = IntensityTable.empty_intensity_table(spot_attributes, 2, 2)
    assert empty.shape == (2, 2, 2)
    assert np.sum(empty.values) == 0


def test_synthetic_intensities_generates_correct_number_of_features(loaded_codebook):
    n_spots = 2
    intensities = IntensityTable.synthetic_intensities(loaded_codebook, n_spots=n_spots)
    assert isinstance(intensities, IntensityTable)

    # shape should have n_spots and channels and hybridization rounds equal to the codebook's shape
    assert intensities.shape == (n_spots, *loaded_codebook.shape[1:])


def test_synthetic_intensities_have_correct_number_of_on_features(loaded_codebook):
    n_spots = 2
    intensities = IntensityTable.synthetic_intensities(loaded_codebook, n_spots=n_spots)
    on_features = np.sum(intensities.values != 0)
    # this asserts that the number of features "on" in intensities should be equal to the
    # number of "on" features in the codewords, times the total number of spots in intensities.
    assert on_features == loaded_codebook.sum((Indices.CH, Indices.HYB)).values.mean() * n_spots



from starfish.image import ImageStack
from starfish.pipeline.algorithmbase import AlgorithmBase


class FilterAlgorithmBase(AlgorithmBase):
    def filter(self, stack: ImageStack) -> None:
        """Performs in-place filtering on an ImageStack."""
        raise NotImplementedError()



import collections
import math
from typing import Tuple, Optional, List

import numpy as np
import pandas as pd
from skimage.measure import regionprops, label
from skimage.measure._regionprops import _RegionProperties
from sklearn.neighbors import NearestNeighbors

from starfish.munge import melt
from starfish.io import Stack
from ._base import PixelFinderAlgorithmBase


class PixelSpotDetector(PixelFinderAlgorithmBase):

    _MerfishDecoderResults = collections.namedtuple(
        '_MerfishDecoderResults',
        ['result', 'decoded_img', 'label_img', 'spot_props'])

    def __init__(
            self, codebook: str, distance_threshold: float=0.5176,
            magnitude_threshold: int=1, area_threshold: int=2, crop_size: int=40, **kwargs) -> None:
        """Decode an image by first coding each pixel, then combining the results into spots

        Parameters
        ----------
        codebook : str
            file mapping codewords to the genes they represent
        distance_threshold : float
            spots whose codewords are more than this distance from an expected code are filtered (default  0.5176)
        magnitude_threshold : int
            spots with intensity less than this value are filtered (default 1)
        area_threshold : int
            spots with total area less than this value are filtered
        crop_size : int
            ??? # TODO ambrosejcarr what is this?

        """
        self.codebook = pd.read_csv(codebook, dtype={'barcode': object})
        self.weighted_codes: pd.DataFrame = self._normalize_barcodes(self.codebook)
        self.distance_threshold = distance_threshold
        self.magnitude_threshold = magnitude_threshold
        self.area_threshold = area_threshold
        self.crop_size = crop_size

        self.label_image: Optional[np.ndarray] = None
        self.decoded_image: Optional[np.ndarray] = None

    @staticmethod
    def encode(stack: Stack) -> pd.DataFrame:

        sq = stack.image.squeeze()
        tile_metadata = stack.image.tile_metadata
        num_bits = int(tile_metadata['barcode_index'].max() + 1)

        # linearize the pixels, mat.shape = (n_hybs * n_channels * n_z_slice, x * y)
        mat = np.reshape(sq.copy(), (sq.shape[0], sq.shape[1] * sq.shape[2]))

        res = pd.DataFrame(mat.T)
        res['spot_id'] = range(len(res))
        res = melt(
            df=res,
            new_index_name='barcode_index',
            new_value_name='intensity',
            melt_columns=range(num_bits)
        )
        # TODO this will be missing spot attributes 'r', because that comes later..., so this can't be attributes, yet
        spots_df_tidy = pd.merge(res, tile_metadata, on='barcode_index', how='left')

        return spots_df_tidy

    def find(self, stack: Stack) -> Tuple[pd.DataFrame, _MerfishDecoderResults]:
        encoded_df: pd.DataFrame = self.encode(stack)
        img_size: Tuple[int, int] = stack.image.tile_shape
        decoded_results = self.decode(encoded=encoded_df, img_size=img_size)

        return encoded_df, decoded_results

    def decode(self, encoded: pd.DataFrame, img_size: Tuple[int, int]) -> _MerfishDecoderResults:
        pixel_traces, pixel_traces_l2_norm = self._parse_pixel_traces(encoded)

        # TODO ambrosejcarr: clean up code in this function
        nn = NearestNeighbors(n_neighbors=1, algorithm='ball_tree').fit(self.weighted_codes)
        distances, indices = nn.kneighbors(pixel_traces)

        # revert back to image space
        decoded_img = np.reshape(indices + 1, img_size)
        decoded_dist = np.reshape(distances, img_size)
        local_magnitude = np.reshape(pixel_traces_l2_norm, img_size)

        # find good 'spots'. filter out bad 'spots'
        decoded_img[decoded_dist > self.distance_threshold] = 0
        decoded_img[local_magnitude < self.magnitude_threshold] = 0
        decoded_img = self._crop(decoded_img, self.crop_size)

        spot_props, label_image, decoded_df = self._find_spots(decoded_img, self.area_threshold)
        decoded_df = pd.merge(decoded_df, self.codebook, on='barcode', how='left')

        # calculate radius of each spot, assuming area is circular
        encoded['radius'] = np.sqrt(decoded_df['area'] / math.pi)
        # encoded[['x, y']] = decoded_df[['x, y']]
        # decoded_df = decoded_df.drop(['x', 'y'], axis=1)
        # spot_attributes = SpotAttributes(encoded)

        self.decoded_image = decoded_img
        self.label_image = label_image

        # DecodedSpots(decoded_df)

        return self._MerfishDecoderResults(decoded_df, decoded_img, label_image, spot_props)

    @staticmethod
    def _normalize_barcodes(codebook: pd.DataFrame) -> np.ndarray:
        # parse barcode into numpy array and normalize by l2_norm
        codes = np.array([np.array([int(d) for d in c]) for c in codebook.barcode])
        codes_l2_norm = np.linalg.norm(codes, axis=1, ord=2)
        weighted_codes = codes / codes_l2_norm[:, None]
        return weighted_codes

    # TODO ambrosejcarr fill in these values + add docs
    @staticmethod
    def _parse_pixel_traces(encoded: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
        # parse spots into pixel traces, normalize and filter
        df = encoded.loc[:, ['spot_id', 'barcode_index', 'intensity']]
        # TODO this assumes that bits are sorted [they are, currently]
        pixel_traces = df.pivot(index='spot_id', columns='barcode_index', values='intensity')
        pixel_traces = pixel_traces.values
        pixel_traces_l2_norm = np.linalg.norm(pixel_traces, axis=1, ord=2)
        ind = pixel_traces_l2_norm > 0
        pixel_traces[ind, :] = pixel_traces[ind, :] / pixel_traces_l2_norm[ind, None]
        return pixel_traces, pixel_traces_l2_norm

    @staticmethod
    def _crop(decoded_img, crop_size) -> np.ndarray:
        decoded_img[:, 0:crop_size] = 0
        decoded_img[:, decoded_img.shape[1] - crop_size:] = 0
        decoded_img[0:crop_size, :] = 0
        decoded_img[decoded_img.shape[0] - crop_size:, :] = 0
        return decoded_img

    def _find_spots(self, decoded_img, area_threshold) -> Tuple[List[_RegionProperties], np.ndarray, pd.DataFrame]:
        label_image = label(decoded_img, connectivity=2)
        props = regionprops(label_image)

        spots = []
        for r in props:
            if r.area >= area_threshold:
                index = decoded_img[int(r.centroid[0]), int(r.centroid[1])]
                if index > 0:
                    data = {'barcode': self.codebook.barcode[index - 1],
                            'x': r.centroid[0],
                            'y': r.centroid[1],
                            'area': r.area
                            }
                    spots.append(data)

        spots_df = pd.DataFrame(spots)
        spots_df['spot_id'] = range(len(spots_df))

        return props, label_image, spots_df

    @classmethod
    def add_arguments(cls, group_parser):
        group_parser.add_argument("--codebook", help="csv file containing a codebook")
        group_parser.add_argument(
            "--distance-threshold", default=0.5176,
            help="maximum distance a pixel may be from a codeword before it is filtered")
        group_parser.add_argument("--magnitude-threshold", type=float, default=1, help="minimum magnitude of a feature")
        group_parser.add_argument("--area-threshold", type=float, default=2, help="minimum area of a feature")
        # TODO ambrosejcarr: figure out help.
        group_parser.add_argument("--crop-size", type=int, default=40, help="???")



import pandas as pd
from ._validated_table import ValidatedTable


class DecodedSpots(ValidatedTable):

    required_fields = {
        # TODO ambrosejcarr change barcode -> codeword, gene -> gene_name (requires rewrite of codebook)
        'barcode',  # the code word that corresponds to this gene
        'quality',  # spot quality
        'spot_id',  # integer spot id
        'gene'  # string gene name
    }

    def __init__(self, decoded_spots: pd.DataFrame) -> None:
        """

        Parameters
        ----------
        decoded_spots : pd.DataFrame

        """
        # call the validation routine, set self.data = required_fields
        super().__init__(decoded_spots, DecodedSpots.required_fields)



from functools import partial

from trackpy import bandpass

from starfish.image import ImageStack
from ._base import FilterAlgorithmBase


class Bandpass(FilterAlgorithmBase):

    def __init__(self, lshort, llong, threshold, truncate, verbose=False, **kwargs):
        """

        Parameters
        ----------
        lshort : float
            filter frequencies below this value
        llong : int
            filter frequencies above this odd integer value
        threshold : float
            zero any pixels below this intensity value
        truncate : float
            # todo document
        verbose : bool
            If True, report on the percentage completed (default = False) during processing
        kwargs
        """
        self.lshort = lshort
        self.llong = llong
        self.threshold = threshold
        self.truncate = truncate
        self.verbose = verbose

    @classmethod
    def add_arguments(cls, group_parser):
        group_parser.add_argument("--lshort", default=0.5, type=float, help="filter signals below this frequency")
        group_parser.add_argument("--llong", default=7, type=int, help="filter signals above this frequency")
        group_parser.add_argument("--threshold", default=1, type=int, help="clip pixels below this intensity value")
        group_parser.add_argument("--truncate", default=4, type=int)

    @staticmethod
    def bandpass(image, lshort, llong, threshold, truncate):
        """Apply a bandpass filter to remove noise and background variation

        Parameters
        ----------
        image : np.ndarray
        lshort : float
            filter frequencies below this value
        llong : int
            filter frequencies above this odd integer value
        threshold : float
            zero any pixels below this intensity value
        truncate : float
            # todo document

        Returns
        -------
        np.ndarray :
            bandpassed image

        """
        bandpassed = bandpass(
            image, lshort=lshort, llong=llong, threshold=threshold,
            truncate=truncate
        )
        return bandpassed

    def filter(self, stack: ImageStack) -> None:
        """Perform in-place filtering of an image stack and all contained aux images.

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.

        """
        bandpass_ = partial(
            self.bandpass, lshort=self.lshort, llong=self.llong, threshold=self.threshold, truncate=self.truncate
        )
        stack.apply(bandpass_, verbose=self.verbose)



import argparse
from functools import partial
from typing import Callable

import numpy as np
from skimage import restoration

from starfish.image import ImageStack
from ._base import FilterAlgorithmBase
from .util import gaussian_kernel


class DeconvolvePSF(FilterAlgorithmBase):

    def __init__(self, num_iter: int=15, sigma: float=2, clip: bool=False, **kwargs) -> None:
        """Deconvolve a point spread function

        Note that the default parameters are highly optimized for the MERFISH use case and that num_iter is a
        very important parameter that requires careful optimization.

        Parameters
        ----------
        num_iter : int
            number of iterations to run
        sigma : float
            standard deviation of the gaussian kernel used to construct the point spread function
        clip : bool (default = False)
            if True, pixel values below -1 and above 1 are clipped for skimage pipeline compatibility

        """
        self.num_iter = num_iter
        self.sigma = sigma
        self.clip = clip
        self.kernel_size: int = int(2 * np.ceil(2 * sigma) + 1)
        self.psf: np.ndarray = gaussian_kernel(shape=(self.kernel_size, self.kernel_size), sigma=sigma)

    @classmethod
    def add_arguments(cls, group_parser: argparse.ArgumentParser) -> None:
        group_parser.add_argument('--num-iter', default=15, type=int, help='number of iterations to run')
        group_parser.add_argument('--sigma', default=2, type=float, help='standard deviation of gaussian kernel')
        group_parser.add_argument(
            '--clip', action='store_true', help='(default False) if True, clip values below -1 and above 1')

    @staticmethod
    def richardson_lucy_deconv(img: np.ndarray, num_iter: int, psf: np.ndarray, clip: bool=False) -> np.ndarray:
        """
        Deconvolves input image with a specified point spread function. This simply calls
        skimage.restoration.richardson_lucy

        Parameters
        ----------
        img : np.ndarray
            Image to filter.
        num_iter : int
            Number of iterations to run algorithm
        psf :
            Point spread function
        clip : bool (default = False)
            If true, pixel value of the result above 1 or under -1 are thresholded for skimage pipeline compatibility.

        Returns
        -------
        np.ndarray :
            Deconvolved image, same shape as input

        """

        # TODO ambrosejcarr: the restoration function is producing the following warning:
        # /usr/local/lib/python3.6/site-packages/skimage/restoration/deconvolution.py:389: RuntimeWarning: invalid value
        # encountered in true_divide:
        # relative_blur = image / convolve_method(im_deconv, psf, 'same')
        img_deconv: np.ndarray = restoration.richardson_lucy(img, psf, iterations=num_iter, clip=clip)

        # here be dragons. img_deconv is a float. this should not work, but the result looks nice
        # modulo boundary values? wtf indeed.
        img_deconv = img_deconv.astype(np.uint16)
        return img_deconv

    def filter(self, stack: ImageStack) -> None:
        """Perform in-place filtering of an image stack and all contained aux images.

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.

        """
        func: Callable = partial(self.richardson_lucy_deconv, num_iter=self.num_iter, psf=self.psf, clip=self.clip)
        stack.apply(func)



import pandas as pd

from starfish.constants import Indices
from ._validated_table import ValidatedTable


class EncodedSpots(ValidatedTable):

    required_fields = {
        'spot_id',  # integer spot id
        'barcode_index',  # position in barcode
        'intensity',  # spot intensity
        Indices.CH.value,  # channel
        Indices.HYB.value,  # hybridization round
    }

    def __init__(self, encoded_spots: pd.DataFrame) -> None:
        """

        Parameters
        ----------
        encoded_spots : pd.DataFrame

        """
        super().__init__(encoded_spots, EncodedSpots.required_fields)



from skimage.feature import blob_log
import pytest
from starfish.pipeline.features.spots.detector.gaussian import GaussianSpotDetector
import numpy as np
# don't inspect pytest fixtures in pycharm
# noinspection PyUnresolvedReferences
from starfish.test.dataset_fixtures import (
    synthetic_dataset_with_truth_values_and_called_spots, synthetic_single_spot_2d,
    synthetic_two_spot_3d, synthetic_single_spot_imagestack_2d, synthetic_single_spot_imagestack_3d,
    synthetic_two_spot_imagestack_3d, synthetic_dataset_with_truth_values, synthetic_single_spot_3d)


def test_spots_match_coordinates_of_synthesized_spots(
        synthetic_dataset_with_truth_values_and_called_spots):

    codebook, true_intensities, image, intensities = (
        synthetic_dataset_with_truth_values_and_called_spots)
    assert true_intensities.shape == intensities.shape

    # we do a bit of rounding, some of these may be off with 1-pixel error.
    x_matches = np.all(np.abs(np.sort(true_intensities.coords['x'].values) -
                              np.sort(intensities.coords['x'].values)) <= 1)
    y_matches = np.all(np.abs(np.sort(true_intensities.coords['y'].values) -
                              np.sort(intensities.coords['y'].values)) <= 1)
    assert x_matches
    assert y_matches


def test_create_intensity_table(synthetic_dataset_with_truth_values_and_called_spots):
    codebook, true_intensities, image, intensities = (
        synthetic_dataset_with_truth_values_and_called_spots)

    assert intensities.shape[0] == 5


def test_create_intensity_table_raises_value_error_when_no_spots_detected(
        synthetic_dataset_with_truth_values_and_called_spots):

    codebook, true_intensities, image, intensities = (
        synthetic_dataset_with_truth_values_and_called_spots)

    min_sigma = 1
    max_sigma = 10
    num_sigma = 30
    threshold = 1000000  # no blobs will be above this high threshold

    gsd = GaussianSpotDetector(
        min_sigma=min_sigma,
        max_sigma=max_sigma,
        num_sigma=num_sigma,
        threshold=threshold,
        blobs_stack=image,
        measurement_type='max',
    )
    with pytest.raises(ValueError):
        gsd.find(hybridization_image=image)


def test_blob_log_2d(synthetic_single_spot_2d):
    result = blob_log(synthetic_single_spot_2d, min_sigma=2, max_sigma=4, num_sigma=10, threshold=0)
    assert np.array_equal(result, np.array([[10, 90, 2]]))


def test_blob_log_3d(synthetic_single_spot_3d):
    result = blob_log(synthetic_single_spot_3d, min_sigma=2, max_sigma=4, num_sigma=10, threshold=0)
    assert np.array_equal(result, np.array([[5, 10, 90, 2]]))


def test_blob_log_3d_flat(synthetic_single_spot_2d):
    """
    verify that 3d blob log works, even when the third dimension is too small to support the
    observed standard deviation
    """
    data = synthetic_single_spot_2d.reshape(1, *synthetic_single_spot_2d.shape)
    result = blob_log(data, min_sigma=2, max_sigma=4, num_sigma=10, threshold=0)
    assert np.array_equal(result, np.array([[0, 10, 90, 2]]))


def single_spot_detector(stack):
    gsd = GaussianSpotDetector(
        min_sigma=1,
        max_sigma=4,
        num_sigma=5,
        threshold=0,
        blobs_stack=stack,
        measurement_type='max',
    )
    return gsd


def test_find_spot_locations_2d(synthetic_single_spot_imagestack_2d):

    gsd = single_spot_detector(synthetic_single_spot_imagestack_2d)
    # noinspection PyProtectedMember
    spot_attributes = gsd._find_spot_locations()

    assert spot_attributes.shape[0] == 1
    assert np.array_equal(spot_attributes[['z', 'y', 'x']].values, np.array([[0, 10, 90]]))
    # rounding incurs an error of up to one pixel
    tol = 1
    assert np.all(
        np.abs(spot_attributes[['z_min', 'z_max', 'y_min', 'y_max', 'x_min', 'x_max']].values -
               np.array([[0, 1, 7, 13, 87, 93]])) <= tol
    )


# noinspection PyProtectedMember
def test_find_spot_locations_3d(synthetic_single_spot_imagestack_3d):
    gsd = single_spot_detector(synthetic_single_spot_imagestack_3d)
    spot_attributes = gsd._find_spot_locations()

    assert spot_attributes.shape[0] == 1
    assert np.array_equal(spot_attributes[['z', 'y', 'x']].values, np.array([[5, 10, 90]]))
    # rounding incurs an error of up to one pixel
    tol = 1
    assert np.all(
        np.abs(spot_attributes[['z_min', 'z_max', 'y_min', 'y_max', 'x_min', 'x_max']].values -
               np.array([[2, 8, 7, 13, 87, 93]])) <= tol
    )


# noinspection PyProtectedMember
def test_measure_spot_intensity_2d(synthetic_single_spot_imagestack_2d):
    gsd = single_spot_detector(synthetic_single_spot_imagestack_2d)
    spot_attributes = gsd._find_spot_locations()
    intensity_table = gsd._measure_spot_intensities(
        synthetic_single_spot_imagestack_2d, spot_attributes)

    assert intensity_table.shape == (1, 1, 1)
    assert np.array_equal(intensity_table.values, np.array([[[np.max(gsd.blobs_image)]]]))


# noinspection PyProtectedMember
def test_measure_spot_intensity_3d(synthetic_single_spot_imagestack_3d):
    gsd = single_spot_detector(synthetic_single_spot_imagestack_3d)
    spot_attributes = gsd._find_spot_locations()
    intensity_table = gsd._measure_spot_intensities(
        synthetic_single_spot_imagestack_3d, spot_attributes)

    assert intensity_table.shape == (1, 1, 1)
    assert np.array_equal(intensity_table.values, np.array([[[np.max(gsd.blobs_image)]]]))


# noinspection PyProtectedMember
def test_measure_two_spots_3d(synthetic_two_spot_imagestack_3d):
    gsd = single_spot_detector(synthetic_two_spot_imagestack_3d)
    spot_attributes = gsd._find_spot_locations()
    intensity_table = gsd._measure_spot_intensities(
        synthetic_two_spot_imagestack_3d, spot_attributes)

    assert intensity_table.shape == (2, 1, 1)
    expected = np.max(gsd.blobs_image).repeat(2).reshape(2, 1, 1)
    assert np.array_equal(intensity_table.values, expected)


def test_find_two_spots_3d(synthetic_two_spot_imagestack_3d):
    gsd = single_spot_detector(synthetic_two_spot_imagestack_3d)
    intensity_table = gsd.find(gsd.blobs_stack)

    assert intensity_table.shape == (2, 1, 1)
    expected = np.max(gsd.blobs_image).repeat(2).reshape(2, 1, 1)
    assert np.array_equal(intensity_table.values, expected)



import argparse

from starfish.pipeline.pipelinecomponent import PipelineComponent
from starfish.util.argparse import FsExistsType
from . import _base
from . import iss


class Decoder(PipelineComponent):

    decoder_group: argparse.ArgumentParser

    @classmethod
    def implementing_algorithms(cls):
        return _base.DecoderAlgorithmBase.__subclasses__()

    @classmethod
    def add_to_parser(cls, subparsers):
        """Adds the decoder component to the CLI argument parser."""
        decoder_group = subparsers.add_parser("decode")
        decoder_group.add_argument("-i", "--input", type=FsExistsType(), required=True)
        decoder_group.add_argument("-o", "--output", required=True)
        decoder_group.add_argument("--codebook", type=FsExistsType(), required=True)
        decoder_group.set_defaults(starfish_command=Decoder._cli)
        decoder_subparsers = decoder_group.add_subparsers(dest="decoder_algorithm_class")

        for algorithm_cls in cls.algorithm_to_class_map().values():
            group_parser = decoder_subparsers.add_parser(algorithm_cls.get_algorithm_name())
            group_parser.set_defaults(decoder_algorithm_class=algorithm_cls)
            algorithm_cls.add_arguments(group_parser)

        cls.decoder_group = decoder_group

    @classmethod
    def _cli(cls, args, print_help=False):
        """Runs the decoder component based on parsed arguments."""
        import pandas

        if args.decoder_algorithm_class is None or print_help:
            cls.decoder_group.print_help()
            cls.decoder_group.exit(status=2)

        instance = args.decoder_algorithm_class(**vars(args))

        encoded = pandas.read_json(args.input, orient="records")
        codebook = pandas.read_json(args.codebook, orient="records")

        results = instance.decode(encoded, codebook)
        print("Writing | spot_id | gene_id to: {}".format(args.output))
        results.save(args.output)



from copy import deepcopy

import numpy as np

from starfish.test.dataset_fixtures import synthetic_stack, labeled_synthetic_dataset


def multiply(array, value):
    return array * value


def test_apply():
    """test that apply correctly applies a simple function across 2d tiles of a Stack"""
    stack = synthetic_stack()
    assert (stack.numpy_array == 1).all()
    stack.apply(multiply, value=2)
    assert (stack.numpy_array == 2).all()


def test_apply_3d():
    """test that apply correctly applies a simple function across 3d volumes of a Stack"""
    stack = synthetic_stack()
    assert np.all(stack.numpy_array == 1)
    stack.apply(multiply, value=4, is_volume=True)
    assert (stack.numpy_array == 4).all()


def test_apply_labeled_dataset():
    """test that apply correctly applies a simple function across starfish-generated synthetic data"""
    original = labeled_synthetic_dataset()
    image = deepcopy(original.image)
    image.apply(multiply, value=2)
    assert np.all(image.numpy_array == original.image.numpy_array * 2)



import numpy as np

from starfish.constants import Indices
from starfish.test.dataset_fixtures import synthetic_stack


def test_get_slice_simple_index():
    """
    Retrieve a slice across one of the indices at the end.  For instance, if the dimensions are
    (P, Q0,..., Qn-1, R), slice across either P or R.
    """
    stack = synthetic_stack()
    hyb = 1
    imageslice, axes = stack.get_slice(
        {Indices.HYB: hyb}
    )
    assert axes == [Indices.CH, Indices.Z]

    Y, X = stack.tile_shape

    for ch in range(stack.shape[Indices.CH]):
        for z in range(stack.shape[Indices.Z]):
            data = np.empty((Y, X))
            data.fill((hyb * stack.shape[Indices.CH] + ch) * stack.shape[Indices.Z] + z)

            assert data.all() == imageslice[ch, z].all()


def test_get_slice_middle_index():
    """
    Retrieve a slice across one of the indices in the middle.  For instance, if the dimensions are
    (P, Q0,..., Qn-1, R), slice across one of the Q axes.
    """
    stack = synthetic_stack()
    ch = 1
    imageslice, axes = stack.get_slice(
        {Indices.CH: ch}
    )
    assert axes == [Indices.HYB, Indices.Z]

    Y, X = stack.tile_shape

    for hyb in range(stack.shape[Indices.HYB]):
        for z in range(stack.shape[Indices.Z]):
            data = np.empty((Y, X))
            data.fill((hyb * stack.shape[Indices.CH] + ch) * stack.shape[Indices.Z] + z)

            assert data.all() == imageslice[hyb, z].all()


def test_get_slice_range():
    """
    Retrieve a slice across a range of one of the dimensions.
    """
    stack = synthetic_stack()
    zrange = slice(1, 3)
    imageslice, axes = stack.get_slice(
        {Indices.Z: zrange}
    )
    Y, X = stack.tile_shape
    assert axes == [Indices.HYB, Indices.CH, Indices.Z]

    for hyb in range(stack.shape[Indices.HYB]):
        for ch in range(stack.shape[Indices.CH]):
            for z in range(zrange.stop - zrange.start):
                data = np.empty((Y, X))
                data.fill((hyb * stack.shape[Indices.CH] + ch) * stack.shape[Indices.Z] +
                          (z + zrange.start))

                assert data.all() == imageslice[hyb, ch, z].all()


def test_set_slice_simple_index():
    """
    Sets a slice across one of the indices at the end.  For instance, if the dimensions are
    (P, Q0,..., Qn-1, R), sets a slice across either P or R.
    """
    stack = synthetic_stack()
    hyb = 1
    Y, X = stack.tile_shape

    expected = np.ones((stack.shape[Indices.CH], stack.shape[Indices.Z], Y, X)) * 2
    index = {Indices.HYB: hyb}

    stack.set_slice(index, expected)

    assert np.array_equal(stack.get_slice(index)[0], expected)


def test_set_slice_middle_index():
    """
    Sets a slice across one of the indices in the middle.  For instance, if the dimensions are
    (P, Q0,..., Qn-1, R), slice across one of the Q axes.
    """
    stack = synthetic_stack()
    ch = 1
    Y, X = stack.tile_shape

    expected = np.ones((stack.shape[Indices.HYB], stack.shape[Indices.Z], Y, X)) * 2
    index = {Indices.CH: ch}

    stack.set_slice(index, expected)

    assert np.array_equal(stack.get_slice(index)[0], expected)


def test_set_slice_range():
    """
    Sets a slice across a range of one of the dimensions.
    """
    stack = synthetic_stack()
    zrange = slice(1, 3)
    Y, X = stack.tile_shape

    expected = np.ones((
        stack.shape[Indices.HYB],
        stack.shape[Indices.CH],
        zrange.stop - zrange.start,
        Y, X)) * 10
    index = {Indices.Z: zrange}

    stack.set_slice(index, expected)

    assert np.array_equal(stack.get_slice(index)[0], expected)



import argparse
import collections
from typing import Mapping, Optional, Type


class PipelineComponentType(type):
    """
    This is the metaclass for PipelineComponent.  As each subclass that is _not_ PipelineComponent is created, it sets
    up a map between the algorithm name and the class that implements it.
    """
    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)
        if len(bases) != 0:
            # this is _not_ PipelineComponent.  Instead, it's a subclass of PipelineComponent.
            PipelineComponentType._ensure_algorithms_setup(cls)

    @classmethod
    def _ensure_algorithms_setup(mcs, cls):
        if cls._algorithm_to_class_map is None:
            cls._algorithm_to_class_map = dict()

            queue = collections.deque(cls.implementing_algorithms())
            while len(queue) > 0:
                algorithm_cls = queue.popleft()
                queue.extend(algorithm_cls.__subclasses__())

                cls._algorithm_to_class_map[algorithm_cls.__name__] = algorithm_cls

                setattr(cls, algorithm_cls.get_algorithm_name(), algorithm_cls)


class PipelineComponent(metaclass=PipelineComponentType):

    _algorithm_to_class_map: Optional[Mapping[str, Type]] = None

    @classmethod
    def implementing_algorithms(cls):
        """
        Get a list of classes that implement an algorithm relevant to this pipeline stage.  Pipeline components must
        provide this method.
        """
        raise NotImplementedError()

    @classmethod
    def algorithm_to_class_map(cls):
        """Returns a mapping from algorithm names to the classes that implement them."""
        return cls._algorithm_to_class_map

    @classmethod
    def run(cls, algorithm_name, stack, *args, **kwargs):
        """Runs the registration component using the algorithm name, stack, and arguments for the specific algorithm."""
        algorithm_cls = cls._algorithm_to_class_map[algorithm_name]
        instance = algorithm_cls(*args, **kwargs)
        return instance.register(stack)

    @classmethod
    def _cli(cls, args: argparse.Namespace):
        raise NotImplementedError()



from typing import Tuple

import numpy as np
from skimage.morphology import disk, binary_opening


def bin_thresh(img: np.ndarray, thresh: int) -> np.ndarray:
    """
    Performs binary thresholding of an image

    Parameters
    ----------
    img : np.ndarray
        Image to filter.
    thresh : int
        Pixel values >= thresh are set to 1, else 0.

    Returns
    -------
    np.ndarray :
        Binarized image, same shape as input

    """
    res = img >= thresh
    return res


def bin_open(img: np.ndarray, disk_size: int) -> np.ndarray:
    """
    Performs binary opening of an image

    img : np.ndarray
        Image to filter.
    disk_size : int
        Radius of the disk-shaped structuring element.

    Returns
    -------
    np.ndarray :
        Filtered image, same shape as input

    """
    selem = disk(disk_size)
    res = binary_opening(img, selem)
    return res


def gaussian_kernel(shape: Tuple[int, int]=(3, 3), sigma: float=0.5):
    """
    Returns a gaussian kernel of specified shape and standard deviation.
    This is a simple python implementation of Matlab's fspecial('gaussian',[shape],[sigma])

    Parameters
    ----------
    shape : Tuple[int] (default = (3, 3))
        Kernel shape.
    sigma : float (default = 0.5)
        Standard deviation of gaussian kernel.

    Parameters
    ----------
    np.ndarray :
        Gaussian kernel.
    """
    m, n = [int((ss - 1.) / 2.) for ss in shape]
    y, x = np.ogrid[-m:m + 1, -n:n + 1]
    kernel = np.exp(-(x * x + y * y) / (2. * sigma * sigma))
    kernel[kernel < np.finfo(kernel.dtype).eps * kernel.max()] = 0
    sumh = kernel.sum()
    if sumh != 0:
        kernel /= sumh
    return kernel



from typing import Any

import pytest

from starfish.constants import Indices
from starfish.test.dataset_fixtures import DEFAULT_NUM_HYB, DEFAULT_NUM_CH, DEFAULT_NUM_Z, synthetic_stack


def test_metadata():
    """
    Normal situation where all the tiles have uniform keys for both indices and extras.
    """
    def tile_extras_provider(hyb: int, ch: int, z: int) -> Any:
        return {
            'random_key': {
                Indices.HYB: hyb,
                Indices.CH: ch,
                Indices.Z: z,
            }
        }

    stack = synthetic_stack(
        tile_extras_provider=tile_extras_provider,
    )
    table = stack.tile_metadata
    assert len(table) == DEFAULT_NUM_HYB * DEFAULT_NUM_CH * DEFAULT_NUM_Z


def test_missing_extras():
    """
    If the extras are not present on some of the tiles, it should still work.
    """
    def tile_extras_provider(hyb: int, ch: int, z: int) -> Any:
        if hyb == 0:
            return {
                'random_key': {
                    Indices.HYB: hyb,
                    Indices.CH: ch,
                    Indices.Z: z,
                }
            }
        else:
            return None

    stack = synthetic_stack(
        tile_extras_provider=tile_extras_provider,
    )
    table = stack.tile_metadata
    assert len(table) == DEFAULT_NUM_HYB * DEFAULT_NUM_CH * DEFAULT_NUM_Z


def test_conflict():
    """
    Tiles that have extras that conflict with indices should produce an error.
    """
    def tile_extras_provider(hyb: int, ch: int, z: int) -> Any:
        return {
            Indices.HYB: hyb,
            Indices.CH: ch,
            Indices.Z: z,
        }

    stack = synthetic_stack(
        tile_extras_provider=tile_extras_provider,
    )
    with pytest.raises(ValueError):
        stack.tile_metadata



import argparse
import json

from starfish.pipeline.pipelinecomponent import PipelineComponent
from starfish.util.argparse import FsExistsType
from . import _base
from . import point_in_poly


class GeneAssignment(PipelineComponent):

    gene_assignment_group: argparse.ArgumentParser

    @classmethod
    def implementing_algorithms(cls):
        return _base.GeneAssignmentAlgorithm.__subclasses__()

    @classmethod
    def add_to_parser(cls, subparsers):
        """Adds the gene_assignment component to the CLI argument parser."""
        gene_assignment_group = subparsers.add_parser("gene_assignment")
        gene_assignment_group.add_argument("--coordinates-geojson", type=FsExistsType(), required=True)
        gene_assignment_group.add_argument("--spots-json", type=FsExistsType(), required=True)
        gene_assignment_group.add_argument("-o", "--output", required=True)
        gene_assignment_group.set_defaults(starfish_command=GeneAssignment._cli)
        gene_assignment_subparsers = gene_assignment_group.add_subparsers(dest="gene_assignment_algorithm_class")

        for algorithm_cls in cls.algorithm_to_class_map().values():
            group_parser = gene_assignment_subparsers.add_parser(algorithm_cls.get_algorithm_name())
            group_parser.set_defaults(gene_assignment_algorithm_class=algorithm_cls)
            algorithm_cls.add_arguments(group_parser)

        cls.gene_assignment_group = gene_assignment_group

    @classmethod
    def _cli(cls, args, print_help=False):
        """Runs the gene_assignment component based on parsed arguments."""
        import pandas
        from starfish import munge

        if args.gene_assignment_algorithm_class is None or print_help:
            cls.gene_assignment_group.print_help()
            cls.gene_assignment_group.exit(status=2)

        with open(args.coordinates_geojson, "r") as fh:
            coordinates = json.load(fh)
        regions = munge.geojson_to_region(coordinates)

        spots = pandas.read_json(args.spots_json, orient="records")

        instance = args.gene_assignment_algorithm_class(**vars(args))

        result = instance.assign_genes(spots, regions)

        print("Writing | cell_id | spot_id to: {}".format(args.output))
        result.to_json(args.output, orient="records")



from ._base import GeneAssignmentAlgorithm


class PointInPoly(GeneAssignmentAlgorithm):
    def __init__(self, **kwargs):
        pass

    @classmethod
    def add_arguments(cls, parser):
        pass

    def assign_genes(self, spots, regions):
        from starfish.assign import assign

        # TODO only works in 3D
        points = spots.loc[:, ['x', 'y']].values
        return assign(regions, points, use_hull=True)



from typing import Any, Callable
from copy import deepcopy

import numpy as np
import pytest
from slicedimage import Tile, TileSet

from starfish.constants import Indices, Coordinates
from starfish.image import ImageStack
from starfish.io import Stack
from starfish.util.synthesize import synthesize


# TODO ambrosejcarr: all fixtures should emit a stack and a codebook
@pytest.fixture(scope='session')
def merfish_stack() -> Stack:
    """retrieve MERFISH testing data from cloudfront and expose it at the module level

    Notes
    -----
    Because download takes time, this fixture runs once per session -- that is, the download is run only once.

    Returns
    -------
    Stack :
        starfish.io.Stack object containing MERFISH data
    """
    s = Stack()
    s.read('https://s3.amazonaws.com/czi.starfish.data.public/20180607/test/MERFISH/fov_001/experiment_new.json')
    return deepcopy(s)


def default_tile_data_provider(hyb: int, ch: int, z: int, height: int, width: int) -> np.ndarray:
    """
    Returns a tile of just ones for any given hyb/ch/z.
    """
    return np.ones((height, width))


def default_tile_extras_provider(hyb: int, ch: int, z: int) -> Any:
    """
    Returns None for extras for any given hyb/ch/z.
    """
    return None


DEFAULT_NUM_HYB = 2
DEFAULT_NUM_CH = 3
DEFAULT_NUM_Z = 4
DEFAULT_HEIGHT = 30
DEFAULT_WIDTH = 20


def synthetic_stack(
        num_hyb: int=DEFAULT_NUM_HYB,
        num_ch: int=DEFAULT_NUM_CH,
        num_z: int=DEFAULT_NUM_Z,
        tile_height: int=DEFAULT_HEIGHT,
        tile_width: int=DEFAULT_WIDTH,
        tile_data_provider: Callable[[int, int, int, int, int], np.ndarray]=default_tile_data_provider,
        tile_extras_provider: Callable[[int, int, int], Any]=default_tile_extras_provider
) -> ImageStack:
    """generate a synthetic ImageStack

    Returns
    -------
    ImageStack :
        imagestack containing a tensor of (2, 3, 4, 30, 20) whose values are all 1.

    """
    img = TileSet(
        {Coordinates.X, Coordinates.Y, Indices.HYB, Indices.CH, Indices.Z},
        {
            Indices.HYB: num_hyb,
            Indices.CH: num_ch,
            Indices.Z: num_z,
        },
        default_tile_shape=(tile_height, tile_width),
    )
    for hyb in range(num_hyb):
        for ch in range(num_ch):
            for z in range(num_z):
                tile = Tile(
                    {
                        Coordinates.X: (0.0, 0.001),
                        Coordinates.Y: (0.0, 0.001),
                        Coordinates.Z: (0.0, 0.001),
                    },
                    {
                        Indices.HYB: hyb,
                        Indices.CH: ch,
                        Indices.Z: z,
                    },
                    extras=tile_extras_provider(hyb, ch, z),
                )
                tile.numpy_array = tile_data_provider(hyb, ch, z, tile_height, tile_width)

                img.add_tile(tile)

    stack = ImageStack(img)
    return stack


def labeled_synthetic_dataset() -> Stack:
    stack, codebook = synthesize()
    return stack
